"""Command line routines related to registering/setting up connections to the Splitgraph registry."""
import hashlib
import logging
import os
import shutil
import string
import subprocess
import sys
from copy import copy
from glob import glob
from pathlib import Path
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, cast
from urllib.parse import quote, urlparse

import click
from click import wrap_text

from splitgraph.cloud.models import AddExternalRepositoryRequest, IntrospectionMode
from splitgraph.cloud.project.models import Metadata, SplitgraphYAML
from splitgraph.commandline.common import (
    ImageType,
    RepositoryType,
    download_file,
    emit_sql_results,
    upload_file,
    wait_for_job,
)
from splitgraph.commandline.engine import inject_config_into_engines
from splitgraph.config.config import get_from_subsection
from splitgraph.config.management import patch_and_save_config
from splitgraph.core.output import Color, pluralise

if TYPE_CHECKING:
    from splitgraph.cloud import GQLAPIClient
    from splitgraph.cloud.project.models import External, Repository
    from splitgraph.core.repository import Repository as CoreRepository

# Hardcoded database name for the Splitgraph DDN (ddn instead of sgregistry)
_DDN_DBNAME = "ddn"

VALID_FILENAME_CHARACTERS = string.ascii_letters + string.digits + "-_."


@click.command("register")
@click.option("--username", prompt=True)
@click.password_option()
@click.option("--email", prompt=True)
@click.option(
    "--remote",
    default="data.splitgraph.com",
    help="Name of the remote registry to register on.",
)
@click.option("--accept-tos", is_flag=True, help="Accept the registry's Terms of Service")
@click.option(
    "-s", "--skip-inject", is_flag=True, help="Don't try to copy the config into all engines"
)
def register_c(username, password, email, remote, accept_tos, skip_inject):
    """
    Register the user on a Splitgraph registry.

    By default, this registers the user on data.splitgraph.com,
    obtains a set of machine (API) credentials for the client to communicate
    with the registry and configures the data.splitgraph.com engine.
    """
    from splitgraph.cloud import DEFAULT_REMOTES, RESTAPIClient
    from splitgraph.config import CONFIG
    from splitgraph.config.config import get_all_in_subsection

    client = RESTAPIClient(remote)
    tos = client.tos()

    if tos and not accept_tos:
        click.echo("%s says: %s" % (client.endpoint, tos))
        click.confirm("Do you accept the Terms of Service?", default=False, abort=True)
    click.echo("Registering the user...")

    uuid, access, refresh = cast(
        Tuple[str, str, str], client.register(username, password, email, accept_tos=True)
    )
    click.echo("Registration successful. UUID %s" % uuid)

    key, secret = cast(Tuple[str, str], client.create_machine_credentials(access, password))
    click.echo("Acquired refresh token and API keys")

    repo_lookup = _update_repo_lookup(CONFIG, remote)

    # If remote isn't in the config (is one of the default ones), add the default parameters
    # to the new config -- otherwise only overwrite the new fields in the existing config.
    remote_params = (
        copy(DEFAULT_REMOTES.get(remote, {}))
        if not get_all_in_subsection(CONFIG, "remotes", remote)
        else {}
    )
    remote_params.update(
        {
            "SG_ENGINE_USER": key,
            "SG_ENGINE_PWD": secret,
            "SG_NAMESPACE": username,
            "SG_CLOUD_REFRESH_TOKEN": refresh,
            "SG_CLOUD_ACCESS_TOKEN": access,
            "SG_IS_REGISTRY": "true",
        }
    )

    config_patch = {
        "SG_REPO_LOOKUP": repo_lookup,
        "remotes": {remote: remote_params},
    }
    config_path = patch_and_save_config(CONFIG, config_patch)

    if not skip_inject:
        inject_config_into_engines(CONFIG["SG_ENGINE_PREFIX"], config_path)

    click.echo("Done.")


def _update_repo_lookup(config, remote):
    repo_lookup = config.get("SG_REPO_LOOKUP")
    if repo_lookup:
        repo_lookup = repo_lookup.split(",")
        if remote not in repo_lookup:
            repo_lookup.append(remote)
    else:
        repo_lookup = [remote]
    return ",".join(repo_lookup)


def _construct_user_profile_url(auth_endpoint: str):
    parsed = urlparse(auth_endpoint)
    netloc_components = parsed.netloc.split(".")
    netloc_components[0] = "www"
    return "%s://%s/settings/security" % (parsed.scheme, ".".join(netloc_components))


def _construct_repo_url(gql_endpoint: str, full_repo: str):
    parsed = urlparse(gql_endpoint)
    netloc_components = parsed.netloc.split(".")
    netloc_components[0] = "www"
    return "%s://%s/%s" % (parsed.scheme, ".".join(netloc_components), full_repo)


def _construct_search_url(gql_endpoint: str, query: str):
    parsed = urlparse(gql_endpoint)
    netloc_components = parsed.netloc.split(".")
    netloc_components[0] = "www"
    return "%s://%s/search?q=%s" % (parsed.scheme, ".".join(netloc_components), quote(query))


@click.command("login")
@click.option("--username", prompt="Username or e-mail")
@click.option("--password", default=None)
@click.option(
    "--remote",
    default="data.splitgraph.com",
    help="Name of the remote registry to log into.",
)
@click.option(
    "--overwrite", is_flag=True, help="Overwrite old API keys in the config if they exist"
)
@click.option(
    "-s", "--skip-inject", is_flag=True, help="Don't try to copy the config into all engines"
)
def login_c(username, password, remote, overwrite, skip_inject):
    """Log into a Splitgraph registry with username/password.

    This will generate a new refresh token (to use the Splitgraph query API)
    and API keys to let sgr access the registry (if they don't already exist
    in the configuration file or if the actual username has changed).

    Note that if you already have generated an API key pair but it's not
    in the configuration file, this will generate a new pair instead of
    restoring the existing one, as the API secret is only stored in the configuration file.

    If you want to log in using an existing API key pair, use `sgr cloud login-api` instead.
    """
    from splitgraph.cloud import DEFAULT_REMOTES, RESTAPIClient, get_token_claim
    from splitgraph.config import CONFIG
    from splitgraph.config.config import get_all_in_subsection

    client = RESTAPIClient(remote)

    if not password:
        profile_url = _construct_user_profile_url(client.endpoint)
        password = click.prompt(
            text="Password (visit %s if you don't have it)" % profile_url,
            confirmation_prompt=False,
            hide_input=True,
        )

    access, refresh = cast(Tuple[str, str], client.get_refresh_token(username, password))

    # Extract namespace from the access token since we might have logged in with an e-mail.
    namespace = get_token_claim(access, "username")

    click.echo("Logged into %s as %s" % (remote, namespace))

    config_remote_params = get_all_in_subsection(CONFIG, "remotes", remote)

    remote_params = copy(DEFAULT_REMOTES.get(remote, {})) if not config_remote_params else {}
    remote_params.update(
        {
            "SG_NAMESPACE": namespace,
            "SG_CLOUD_REFRESH_TOKEN": refresh,
            "SG_CLOUD_ACCESS_TOKEN": access,
        }
    )

    config_patch = {
        "SG_REPO_LOOKUP": _update_repo_lookup(CONFIG, remote),
        "remotes": {remote: remote_params},
    }

    # Get new tokens in any case if we're logging in under a different username.
    try:
        username_changed = namespace != get_from_subsection(
            CONFIG, "remotes", remote, "SG_NAMESPACE"
        )
    except KeyError:
        username_changed = False

    if (
        "SG_ENGINE_USER" not in config_remote_params
        or "SG_ENGINE_PWD" not in config_remote_params
        or overwrite
        or username_changed
    ):
        key, secret = cast(Tuple[str, str], client.create_machine_credentials(access, password))
        config_patch["remotes"][remote]["SG_ENGINE_USER"] = key
        config_patch["remotes"][remote]["SG_ENGINE_PWD"] = secret
        click.echo("Acquired new API keys")

    config_path = patch_and_save_config(CONFIG, config_patch)

    if not skip_inject:
        inject_config_into_engines(CONFIG["SG_ENGINE_PREFIX"], config_path)


@click.command("login-api")
@click.option("--api-key", prompt="API key")
@click.option("--api-secret", prompt="API secret", confirmation_prompt=False, hide_input=True)
@click.option(
    "--remote",
    default="data.splitgraph.com",
    help="Name of the remote registry to log into.",
)
@click.option(
    "-s", "--skip-inject", is_flag=True, help="Don't try to copy the config into all engines"
)
def login_api_c(api_key, api_secret, remote, skip_inject):
    """Log into a Splitgraph registry using existing API keys.

    This will inject the API keys for the registry into the configuration file
    and generate a new access token.
    """
    from splitgraph.cloud import DEFAULT_REMOTES, RESTAPIClient, get_token_claim
    from splitgraph.config import CONFIG
    from splitgraph.config.config import get_all_in_subsection

    client = RESTAPIClient(remote)
    access = cast(str, client.get_access_token_from_api(api_key, api_secret))

    namespace = get_token_claim(access, "username")

    click.echo("Logged into %s as %s" % (remote, namespace))
    remote_params = (
        copy(DEFAULT_REMOTES.get(remote, {}))
        if not get_all_in_subsection(CONFIG, "remotes", remote)
        else {}
    )
    remote_params.update(
        {
            "SG_NAMESPACE": namespace,
            "SG_CLOUD_ACCESS_TOKEN": access,
            "SG_ENGINE_USER": api_key,
            "SG_ENGINE_PWD": api_secret,
        }
    )
    config_patch = {
        "SG_REPO_LOOKUP": _update_repo_lookup(CONFIG, remote),
        "remotes": {remote: remote_params},
    }
    config_path = patch_and_save_config(CONFIG, config_patch)

    if not skip_inject:
        inject_config_into_engines(CONFIG["SG_ENGINE_PREFIX"], config_path)


@click.command("curl", context_settings={"ignore_unknown_options": True})
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option(
    "-t", "--request-type", default="postgrest", type=click.Choice(["postgrest", "splitfile"])
)
@click.argument("image", type=ImageType(default="latest"))
@click.argument("request_params", type=str, default="")
# This is kind of awkward: we want to support not passing in any request params to default
# them to / but also want to be able to support multiple curl args which click doesn't let
# us do (click.argument with nargs=-1 makes "request_params" consume the first arg even
# if we use the posix standard separator --).
@click.option(
    "-c", "--curl-args", type=str, multiple=True, help="Extra arguments to be passed to curl"
)
def curl_c(remote, request_type, image, request_params, curl_args):
    """
    Query a Splitgraph REST API.

    This is a thin wrapper around curl that performs an HTTP request to Splitgraph Cloud to
    interact with a dataset using PostgREST (http://postgrest.org) or the Splitfile execution service.

    The actual invocation is:

    ```
    curl [API endpoint][request] -H [access_token] [extra curl args].
    ```

    The image must be of the form `namespace/repository:[hash_or_tag (default latest)]`.

    The actual request parameters depend on the request type:

      * For PostgREST: `/table?[postgrest request]` or empty to get the OpenAPI spec for this image.
        For a reference on how to perform Postgrest requests, see http://postgrest.org/en/latest/api.html.
      * For the Splitfile executor: a JSON array to be POSTed to the executor, e.g.
        `'{"command": "FROM some/repo IMPORT some_table AS alias", "tag": "new_tag"}'`.

    `--curl-args` allows to pass extra arguments to curl. Note that every argument must be prefixed
    with `--curl-args`, e.g. `--curl-args --cacert --curl-args /path/to/ca.pem`.
    """
    from splitgraph.cloud import RESTAPIClient, get_headers
    from splitgraph.config import CONFIG

    repository, hash_or_tag = image

    # Craft a request
    config = cast(Dict[str, str], CONFIG["remotes"][remote])
    access_token = RESTAPIClient(remote).access_token
    headers = get_headers()
    headers.update({"Authorization": "Bearer " + access_token})

    if request_type == "postgrest":
        if request_params and not request_params.startswith("/"):
            request_params = "/" + request_params
        full_request = (
            config["SG_QUERY_API"]
            + "/%s/%s" % (str(repository), str(hash_or_tag))
            + "/-/rest"
            + request_params
        )
    else:
        full_request = (
            config["SG_QUERY_API"] + "/%s/%s" % (str(repository), str(hash_or_tag)) + "/-/splitfile"
        )
        curl_args = ["-X", "POST", "-d", request_params] + list(curl_args)
        headers.update({"Content-Type": "application/json"})

    header_invocation = [h for i in headers.items() for h in ("-H", "%s: %s" % i)]
    subprocess_args = ["curl", full_request] + header_invocation + list(curl_args)

    logging.debug("Calling %s", " ".join(subprocess_args))
    subprocess.call(subprocess_args)


def _get_ddn_conn_params(remote: str) -> Dict[str, Optional[str]]:
    from splitgraph.engine import get_engine

    try:
        engine = get_engine(remote)
    except KeyError as e:
        raise click.UsageError(
            "Remote %s or API key/secret not found in the config. "
            "Try registering with sgr cloud register or logging in "
            "with sgr cloud login / sgr cloud login-api."
        ) from e

    ddn_params = engine.conn_params.copy()
    ddn_params["SG_ENGINE_DB_NAME"] = _DDN_DBNAME
    return ddn_params


@click.command("sql")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option("-a", "--show-all", is_flag=True, help="Return all results of the query.")
@click.option("-j", "--json", is_flag=True, help="Return results as JSON")
@click.argument("query", type=str, default="")
def sql_c(remote, show_all, json, query):
    """
    Run SQL on or connect to the Splitgraph Data Delivery Network.

    If a query isn't passed, this will return a libpq-compatible connection string to
    the registry's SQL endpoint. It can be used to connect to the endpoint with other SQL clients:

    ```
    pgcli $(sgr cloud sql)
    ```

    If a query is passed, this will run an SQL query against the SQL endpoint.
    """
    ddn_params = _get_ddn_conn_params(remote)
    from splitgraph.engine.postgres.engine import PostgresEngine, get_conn_str

    if not query:
        click.echo(get_conn_str(ddn_params))
        return

    # Build an engine to connect to the DDN, disable pre-flight API checks etc
    engine = PostgresEngine(
        name=remote, conn_params=ddn_params, registry=False, check_version=False
    )

    try:
        results = engine.run_sql(query)
        emit_sql_results(results, use_json=json, show_all=show_all)
    finally:
        engine.close()


@click.command("readme")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.argument("repository", type=RepositoryType(exists=False))
@click.argument("readme", type=click.File("r"))
def readme_c(remote, repository, readme):
    """Upload a README to a Splitgraph repository.

    The repository must have already been pushed. The README must be a file in Markdown format.
    """
    from splitgraph.cloud import GQLAPIClient

    client = GQLAPIClient(remote)
    client.upsert_readme(
        namespace=repository.namespace, repository=repository.repository, readme=readme.read()
    )
    click.echo("README updated for repository %s." % str(repository))


@click.command("description")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.argument("repository", type=RepositoryType(exists=False))
@click.argument("description", type=str)
def description_c(remote, repository, description):
    """Upload a description to a Splitgraph repository.

    The repository must have already been pushed. The description should be plain text, 160 characters or shorter.
    """
    from splitgraph.cloud import GQLAPIClient

    client = GQLAPIClient(remote)
    client.upsert_description(
        namespace=repository.namespace, repository=repository.repository, description=description
    )
    click.echo("Description updated for repository %s." % str(repository))


@click.command("metadata")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.argument("repository", type=RepositoryType(exists=False))
@click.argument("metadata_file", type=click.File("r"), default="splitgraph.yml")
@click.pass_context
def metadata_c(ctx, remote, repository, metadata_file):
    """Upload a metadata file to a Splitgraph repository.

    This can manipulate the repository's short description, README, topics, license, sources and extra metadata.

    The metadata file must be a YAML file. Omitting a key doesn't delete the value.

    For example:

    \b
    ```
    readme: dataset-readme.md
    description: Dataset description (160 characters max).
    topics:
      - topic_1
      - topic_2
    sources:
      - anchor: Source
        href: https://www.splitgraph.com
        isCreator: true
        isSameAs: false
      - anchor: Source 2
        href: https://www.splitgraph.com
        isCreator: false
        isSameAs: true
    license: Public Domain
    extra_metadata:
      key_1:
        key_1_1: value_1_1
        key_1_2: value_1_2
      key_2:
        key_2_1: value_2_1
        key_2_2: value_2_2
    ```
    """
    from splitgraph.cloud import GQLAPIClient
    from splitgraph.utils.yaml import safe_load

    metadata = Metadata.parse_obj(safe_load(metadata_file))
    metadata = _prepare_metadata(metadata)

    client = GQLAPIClient(remote)
    client.upsert_metadata(repository.namespace, repository.repository, metadata)
    click.echo("Metadata updated for repository %s." % str(repository))


def _prepare_metadata(metadata, readme_basedir="."):
    keys = ["readme", "description", "topics", "sources", "license", "extra_metadata"]
    if all(metadata.__getattribute__(k) is None for k in keys):
        raise click.UsageError(
            "Invalid metadata file. File must contain at least one of " f"{'/'.join(keys)} keys."
        )
    readme_file = None
    if isinstance(metadata.readme, str):
        readme_file = metadata.readme
    elif isinstance(metadata.readme, Metadata.Readme) and metadata.readme.file:
        readme_file = metadata.readme.file
    if readme_file:
        with open(os.path.join(readme_basedir, readme_file), "r") as f:
            metadata.readme = Metadata.Readme(text=f.read())

    return metadata


@click.command("search")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option(
    "--limit", default=10, type=click.IntRange(10, 30), help="Number of results to return"
)
@click.argument("query", type=str)
def search_c(remote, query, limit=10):
    """
    Search for a repository on the Splitgraph registry.

    For more advanced search, including filtering by topics, go to the registry's website itself.
    """

    from splitgraph.cloud import GQLAPIClient

    client = GQLAPIClient(remote)
    total_count, repos_previews = client.find_repository(query=query, limit=limit)

    terminal_width, _ = shutil.get_terminal_size((80, 25))

    for ns, repo, preview in repos_previews:
        full_repo = ns + "/" + repo
        click.echo(Color.BOLD + full_repo + Color.END)
        click.echo(Color.BLUE + _construct_repo_url(client.endpoint, full_repo) + Color.END)
        click.echo(
            wrap_text(
                preview.replace("<<", Color.BOLD).replace(">>", Color.END),
                width=min(terminal_width, 100),
                initial_indent="  ",
                subsequent_indent="  ",
                preserve_paragraphs=True,
            )
        )
        click.echo("\n")

    click.echo("\nTotal results: %d" % total_count)
    if len(repos_previews) < total_count:
        click.echo(
            "Visit %s for more search options and full results."
            % (Color.BLUE + _construct_search_url(client.endpoint, query) + Color.END)
        )


def _normalise_filename(filename):
    """
    Strips out odd characters from a string so that it can be used as a valid filename.
    In order for the result to be unique, we also make a short hash of the original
    string and stick it on the end
    """
    base = "".join(c for c in filename if c in VALID_FILENAME_CHARACTERS)
    short_hash = hashlib.sha1(filename.encode()).hexdigest()[:4]  # nosec
    return f"{base}.{short_hash}"


@click.command("dump")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option("--readme-dir", default="./readmes", help="Directory to dump the data into")
@click.option("--repositories-file", "-f", default="splitgraph.yml", type=click.File("w"))
@click.argument("limit_repositories", type=str, nargs=-1)
def dump_c(remote, readme_dir, repositories_file, limit_repositories):
    """
    Dump a Splitgraph catalog to a YAML file.

    This creates a splitgraph.yml file in the target directory as well as a subdirectory `readmes`
    with all the repository readmes. This file can be used to recreate all catalog metadata
    and all external data source settings for a repository using `sgr cloud load`.
    """
    from splitgraph.cloud import GQLAPIClient
    from splitgraph.cloud.project.utils import dump_project

    client = GQLAPIClient(remote)
    repositories = client.load_all_repositories(limit_to=limit_repositories)

    _dump_readmes_to_dir(repositories, readme_dir)
    dump_project(SplitgraphYAML(repositories=repositories), repositories_file)


@click.command("load")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option(
    "--readme-dir", default="./readmes", help="Path to the directory with the README files"
)
@click.option(
    "-p",
    "--initial-private",
    is_flag=True,
    help="If the repository doesn't exist, create it as private.",
)
@click.option(
    "--repositories-file", "-f", default=["splitgraph.yml"], type=click.Path(), multiple=True
)
@click.option(
    "--skip-external",
    is_flag=True,
    help="Only set up the metadata, not the external data source settings",
)
@click.option(
    "--introspection-mode",
    type=click.Choice(IntrospectionMode),
    default=IntrospectionMode.EMPTY,
    help="Whether to reintrospect tables. none: never reintrospect. all: reintrospect all tables. "
    "empty: only reintrospect tables with an empty schema.",
)
@click.option(
    "--ignore-introspection-errors",
    is_flag=True,
    help="If set, will ignore errors when introspecting tables.",
)
@click.argument("limit_repositories", type=str, nargs=-1)
def load_c(
    remote,
    readme_dir,
    skip_external,
    initial_private,
    repositories_file,
    limit_repositories,
    introspection_mode,
    ignore_introspection_errors,
):
    """
    Load a Splitgraph catalog from a YAML file.

    This will load a splitgraph.yml file and the `readmes` subdirectory produced by
    `sgr cloud dump` back into a remote Splitgraph catalog.

    By default, if a repository doesn't yet exist, it will be public. Pass `--initial-private` to
    this command to make it start off as private. This won't apply to existing repositories: use
    the Splitgraph GUI to manage repository access settings.
    """
    from splitgraph.cloud import GQLAPIClient, RESTAPIClient
    from splitgraph.cloud.project.utils import load_project

    repo_yaml = load_project(repositories_file)
    repositories = repo_yaml.repositories
    if limit_repositories:
        repositories = [
            r for r in repositories if f"{r.namespace}/{r.repository}" in limit_repositories
        ]

    gql_client = GQLAPIClient(remote)

    if not skip_external:
        rest_client = RESTAPIClient(remote)

        filter_credential_names = [
            r.external.credential for r in repositories if r.external and r.external.credential
        ]

        # Set up and load credential IDs from the remote to allow users to refer to them by ID
        # or by a name. Only include credentials for repositories that we're actually loading.
        credential_map = _build_credential_map(
            rest_client,
            credentials=repo_yaml.credentials or {},
            filter_credential_names=filter_credential_names,
        )

        logging.info("Uploading images...")
        external_repositories = []
        for repository in repositories:
            if repository.external:
                external_repository = AddExternalRepositoryRequest.from_external(
                    repository.namespace,
                    repository.repository,
                    repository.external,
                    credential_map=credential_map,
                    initial_private=initial_private,
                )
                external_repositories.append(external_repository)
        rest_client.bulk_upsert_external(
            repositories=external_repositories,
            introspection_mode=introspection_mode,
            raise_errors=not ignore_introspection_errors,
        )
        logging.info(f"Uploaded images for {pluralise('repository', len(external_repositories))}")

    logging.info("Updating metadata...")
    namespace_list = []
    repository_list = []
    metadata_list = []
    for repository in repositories:
        if repository.metadata:
            namespace_list.append(repository.namespace)
            repository_list.append(repository.repository)

            metadata = _prepare_metadata(repository.metadata, readme_basedir=readme_dir)
            metadata_list.append(metadata)
    gql_client.bulk_upsert_metadata(namespace_list, repository_list, metadata_list)
    logging.info(f"Updated metadata for {pluralise('repository', len(repository_list))}")


def _build_credential_map(
    auth_client, credentials=None, filter_credential_names: Optional[List[str]] = None
):
    credential_map = {}
    if credentials:
        logging.info("Setting up credentials on the remote...")
        for credential_name, credential in credentials.items():
            if filter_credential_names and credential_name not in filter_credential_names:
                continue
            credential_id = auth_client.ensure_external_credential(
                credential_data=credential.data,
                credential_name=credential_name,
                plugin_name=credential.plugin,
            )
            logging.info("%s: %s", credential_name, credential_id)
            credential_map[credential_name] = credential_id

    # Load any other credentials too (so that we can use existing credentials by
    # name in YAML files without redefining them)
    response = auth_client.list_external_credentials()
    for credential in response.credentials:
        if credential.credential_name not in credential_map and (
            not filter_credential_names or credential.credential_name in filter_credential_names
        ):
            credential_map[credential.credential_name] = credential.credential_id

    return credential_map


def _dump_readmes_to_dir(repositories: List["Repository"], readme_dir: str) -> None:
    """
    READMEs aren't rendering very nicely in YAML so we instead write them out to
    individual files and replace the README contents with their file paths.

    :param repositories: will be modified by this function
    """
    try:
        os.mkdir(readme_dir)
    except FileExistsError:
        logging.info("Directory %s already exists, cleaning out", readme_dir)
        for path in glob(os.path.join(readme_dir, "*.md")):
            os.unlink(path)
    for repo in repositories:
        if not repo.metadata:
            continue

        readme_text = None
        if isinstance(repo.metadata.readme, str):
            readme_text = repo.metadata.readme
        if (
            isinstance(repo.metadata.readme, Metadata.Readme)
            and repo.metadata.readme.file is None
            and repo.metadata.readme.text is not None
        ):
            readme_text = repo.metadata.readme.text

        if readme_text:
            filename = f"{repo.namespace}-{repo.repository}"
            filename = _normalise_filename(filename)
            filename = f"{filename}.md"
            logging.info("Writing %s", filename)
            path = os.path.join(readme_dir, filename)
            with open(path, "w") as f:
                f.write(readme_text)
            repo.metadata.readme = Metadata.Readme(file=filename)


@click.command("token")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
def token_c(remote):
    """Output an up-to-date access token for a remote."""
    from splitgraph.cloud import RESTAPIClient

    client = RESTAPIClient(remote)
    click.echo(client.access_token)


@click.command("add")
@click.option("--remote", help="Name for the remote (infer from the domain name by default)")
@click.option(
    "-s", "--skip-inject", is_flag=True, help="Don't try to copy the config into all engines"
)
@click.argument("domain_name")
def add_c(remote, skip_inject, domain_name):
    """Add a remote Splitgraph registry to .sgconfig with default parameters"""
    from splitgraph.config import CONFIG
    from splitgraph.config.config import get_all_in_subsection

    if domain_name.startswith(("data", "www", "api")):
        raise click.BadArgumentUsage(
            "Use the base domain name, e.g. " "splitgraph.com instead of data.splitgraph.com"
        )

    if not remote:
        remote = domain_name.split(".")[0]

    remote_params_patch: Dict[str, str] = {
        "SG_IS_REGISTRY": "true",
        "SG_ENGINE_HOST": f"data.{domain_name}",
        "SG_ENGINE_PORT": "5432",
        "SG_ENGINE_DB_NAME": "sgregistry",
        "SG_AUTH_API": f"https://api.{domain_name}/auth",
        "SG_QUERY_API": f"https://data.{domain_name}",
        "SG_GQL_API": f"https://api.{domain_name}/gql/cloud/unified/graphql",
    }

    click.echo("Adding remote %s to the config. Parameters: %s" % (remote, remote_params_patch))

    remote_params = get_all_in_subsection(CONFIG, "remotes", remote) or {}
    remote_params.update(remote_params_patch)

    config_patch = {
        "SG_REPO_LOOKUP": _update_repo_lookup(CONFIG, remote),
        "remotes": {remote: remote_params},
    }
    config_path = patch_and_save_config(CONFIG, config_patch)

    if not skip_inject:
        inject_config_into_engines(CONFIG["SG_ENGINE_PREFIX"], config_path)

    click.echo(
        "Done. You can now register or "
        "log in to the instance with sgr cloud register/login(-api)."
    )


@click.command("status")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option(
    "--repositories-file", "-f", default=["splitgraph.yml"], type=click.Path(), multiple=True
)
@click.argument("repositories", type=RepositoryType(exists=False), nargs=-1)
def status_c(remote, repositories_file, repositories):
    """
    Get job statuses for given repositories.

    If this command is passed a list of repositories, it will get the latest job
    statuses just for those repositories. Otherwise, it will use an existing splitgraph.yml
    file and get job statuses for all repositories in this file.
    """
    from tabulate import tabulate

    from splitgraph.cloud import GQLAPIClient
    from splitgraph.cloud.project.utils import load_project

    client = GQLAPIClient(remote)

    if repositories:
        repo_list = [(r.namespace, r.repository) for r in repositories]
    else:
        repo_yaml = load_project(repositories_file)
        repo_list = [(r.namespace, r.repository) for r in repo_yaml.repositories]

    table: List[Tuple] = []
    for namespace, repository in repo_list:
        job_status = client.get_latest_ingestion_job_status(namespace, repository)
        if job_status:
            table.append(
                (
                    namespace + "/" + repository,
                    job_status.task_id,
                    job_status.started,
                    job_status.finished,
                    job_status.is_manual,
                    job_status.status,
                )
            )
        else:
            table.append((namespace + "/" + repository, None, None, None, None, None))

    click.echo(
        tabulate(
            table,
            headers=[
                "Repository",
                "Task ID",
                "Started",
                "Finished",
                "Manual",
                "Status",
            ],
        )
    )


@click.command("logs")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.argument("repository", type=RepositoryType(exists=False))
@click.argument("task_id", type=str)
def logs_c(remote, repository, task_id):
    """
    Get ingestion job logs.

    Output the logs for an ingestion job into a repository on Splitgraph Cloud.
    Use `sgr cloud status` to get a list of ingestion jobs and their recent task IDs.
    """
    from splitgraph.cloud import GQLAPIClient

    client = GQLAPIClient(remote)

    logs = client.get_ingestion_job_logs(
        namespace=repository.namespace, repository=repository.repository, task_id=task_id
    )
    click.echo(logs)


def _deduplicate_items(items: List[str]) -> List[str]:
    """
    Deduplicate items in a list by adding a numerical suffix to duplicates.
    """
    counts: Dict[str, int] = {}
    positions: List[Tuple[str, int]] = []
    for item in items:
        count = counts.get(item, 0)
        positions.append((item, count))
        counts[item] = count + 1

    result: List[str] = []
    for (item, position) in positions:
        if counts[item] > 1:
            result.append(f"{item}_{position:03d}")
        else:
            result.append(item)
    return result


@click.command("upload")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option("--file-format", default="csv", type=click.Choice(["csv"]))
@click.argument("repository", type=RepositoryType(exists=False))
@click.argument("files", type=click.File("rb"), nargs=-1)
def upload_c(remote, file_format, repository, files):
    """
    Upload files to Splitgraph.

    This uses the upload API to add data like CSV files to a remote Splitgraph instance,
    trigger a load and wait for the data to load into a repository.
    """

    from splitgraph.cloud import GQLAPIClient

    client = GQLAPIClient(remote)

    click.echo("Uploading the files...")

    table_names = _deduplicate_items(
        [os.path.splitext(os.path.basename(f.name))[0].lower() for f in files]
    )

    download_urls = []
    for file in files:
        upload, download = client.get_csv_upload_download_urls()
        download_urls.append(download)
        upload_file(file, upload)

    task_id = client.start_csv_load(
        repository.namespace, repository.repository, download_urls, table_names
    )

    wait_for_load(client, repository.namespace, repository.repository, task_id)

    web_url = _construct_repo_url(gql_endpoint=client.endpoint, full_repo=repository) + "/-/tables"
    click.echo()
    click.echo(
        "Success. See the repository at " + Color.BLUE + web_url + Color.END + " or query it with:"
    )
    click.echo(f'    sgr cloud sql \'SELECT * FROM "{repository}"."{table_names[0]}"\'')  # nosec


def wait_for_load(client: "GQLAPIClient", namespace: str, repository: str, task_id: str) -> None:
    final_status = wait_for_job(
        task_id, lambda: client.get_latest_ingestion_job_status(namespace, repository)
    )

    if final_status.status == "FAILURE":
        logs = client.get_ingestion_job_logs(
            namespace=namespace, repository=repository, task_id=task_id
        )
        click.echo(logs)
        raise ValueError("Error loading data")


def wait_for_download(client: "GQLAPIClient", task_id: str) -> str:
    final_status = wait_for_job(task_id, lambda: client.get_export_job_status(task_id))
    if final_status.status == "SUCCESS":
        assert final_status.output
        return str(final_status.output["url"])
    else:
        raise ValueError(
            "Error running query. This could be due to a syntax error. "
            "Run the query interactively with `sgr cloud sql` to investigate the cause."
        )


@click.command("download")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option("--file-format", default="csv", type=click.Choice(["csv"]))
@click.argument("query", type=str)
@click.argument("output_path", type=str, default=None, required=False)
def download_c(remote, file_format, query, output_path):
    """
    Download query results from Splitgraph.

    This runs a query on Splitgraph Cloud and exports the results into a csv.gz format.
    """
    from splitgraph.cloud import GQLAPIClient

    client = GQLAPIClient(remote)

    task_id = client.start_export(query=query)
    download_url = wait_for_download(client, task_id)
    download_file(download_url, output_path)


@click.command("sync")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option(
    "-r",
    "--full-refresh",
    help="Trigger a full reload instead of an incremental load",
    is_flag=True,
)
@click.option("-w", "--wait", help="Attach to the job and wait for it to finish", is_flag=True)
@click.option(
    "-p",
    "--initial-private",
    is_flag=True,
    help="If the repository doesn't exist, create it as private.",
)
@click.option("-u", "--use-file", is_flag=True, help="Use a YAML file with repository settings")
@click.option(
    "-f", "--repositories-file", default=["splitgraph.yml"], type=click.Path(), multiple=True
)
@click.argument("repository", type=RepositoryType(exists=False))
def sync_c(remote, full_refresh, wait, initial_private, use_file, repositories_file, repository):
    """
    Trigger an ingestion job for a repository.

    This starts off a load job for an existing/new repository, optionally using a
    splitgraph.yml file and waiting for the job to complete before exiting.

    If the splitgraph.yml file is specified, it will use the settings for that repository
    from there to override the existing parameters for a repository or to create a new repository.
    Otherwise, it will use the existing parameters.

    By default, if a repository doesn't yet exist, it will be public. Pass `--initial-private` to
    this command to make it start off as private. This won't apply to existing repositories: use
    the Splitgraph GUI to manage repository access settings.
    """
    from splitgraph.cloud import GQLAPIClient

    client = GQLAPIClient(remote)

    if not use_file:
        if initial_private:
            logging.warning(
                "--initial-private is ignored when manipulating existing repositories. "
                "Use the Splitgraph GUI to manage repository access settings."
            )
        task_id = client.start_load_existing(
            namespace=repository.namespace, repository=repository.repository, sync=not full_refresh
        )
    else:
        external, credential_data = _get_external_from_yaml(repositories_file, repository)

        task_id = client.start_load_params(
            namespace=repository.namespace,
            repository=repository.repository,
            external=external,
            sync=not full_refresh,
            credential_data=credential_data,
            initial_private=initial_private,
        )

    if not wait:
        click.echo("Started task %s" % task_id)
    else:
        wait_for_load(client, repository.namespace, repository.repository, task_id)


def _get_external_from_yaml(
    repositories_file: List[Path], repository: "CoreRepository"
) -> Tuple["External", Optional[Dict[str, Any]]]:
    from splitgraph.cloud.project.utils import load_project

    repo_yaml = load_project(repositories_file)

    for repo_settings in repo_yaml.repositories:
        if (
            repo_settings.namespace == repository.namespace
            and repo_settings.repository == repository.repository
        ):
            break
    else:
        raise click.UsageError(
            "Repository %s not found in the splitgraph.yml file" % repository.to_schema()
        )
    if not repo_settings.external:
        raise click.UsageError(
            "Repository %s doesn't have external settings" % repository.to_schema()
        )
    credential_data = None
    if repo_settings.external.credential:
        if (
            not repo_yaml.credentials
            or repo_settings.external.credential not in repo_yaml.credentials
        ):
            raise click.UsageError(
                "Credential %s not defined in splitgraph.yml" % repo_settings.external.credential
            )
        credential_data = repo_yaml.credentials[repo_settings.external.credential].data
    return repo_settings.external, credential_data


@click.command("plugins")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option(
    "-f",
    "--filter",
    "filter_plugins",
    help="Filter for plugins with this string in name/description",
    type=str,
)
def plugins_c(remote, filter_plugins):
    """List all Splitgraph Cloud data plugins.

    This command lists all plugins available on a remote Splitgraph Cloud instance.
    """

    import tabulate

    from splitgraph.cloud import GQLAPIClient

    client = GQLAPIClient(remote)

    plugins = client.get_all_plugins()
    if filter_plugins:
        plugins = [
            p
            for p in plugins
            if filter_plugins.lower() in " ".join([p.name, p.plugin_name, p.description]).lower()
        ]

    plugins = sorted(plugins, key=lambda p: p.name)
    click.echo(
        tabulate.tabulate(
            [(p.plugin_name, p.name, p.description) for p in plugins],
            headers=["ID", "Name", "Description"],
        )
    )


@click.command("stub")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.argument("plugin_name", type=str)
@click.argument("repository", type=RepositoryType(exists=False))
@click.argument("output_file", type=click.File("w"), default="-")
def stub_c(remote, plugin_name, repository, output_file):
    """Generate a splitgraph.yml stub file for a given plugin"""
    import ruamel.yaml

    from splitgraph.cloud import GQLAPIClient
    from splitgraph.cloud.project.generation import stub_plugin

    client = GQLAPIClient(remote)

    plugin = client.get_plugin(plugin_name)
    if not plugin:
        raise click.UsageError("Plugin %s not found on remote %s" % (plugin_name, remote))

    yml = ruamel.yaml.YAML()
    output = stub_plugin(plugin, repository.namespace, repository.repository)
    yml.dump(output, output_file)


@click.command("validate")
@click.option(
    "-f", "--repositories-file", default=["splitgraph.yml"], type=click.Path(), multiple=True
)
def validate_c(repositories_file):
    """Validate, merge and output project file(s)"""
    from splitgraph.cloud.project.utils import dump_project, load_project

    project = load_project(repositories_file)
    dump_project(project, sys.stdout)


@click.command("seed")
@click.option("--remote", default="data.splitgraph.com", help="Name of the remote registry to use.")
@click.option("--github-repository", help="Name of the GitHub repository")
@click.argument("seed")
@click.argument("directory", type=click.Path(file_okay=False), default=".")
def seed_c(remote, seed, github_repository, directory):
    """Generate a starter Splitgraph Cloud project from a seed."""
    from splitgraph.cloud import GQLAPIClient
    from splitgraph.cloud.project.generation import generate_project

    client = GQLAPIClient(remote)

    generate_project(client, seed, directory, github_repo=github_repository)
    click.echo(f"Splitgraph project generated in {os.path.abspath(directory)}.")


@click.group("cloud")
def cloud_c():
    """Run actions on Splitgraph Cloud."""


cloud_c.add_command(login_c)
cloud_c.add_command(login_api_c)
cloud_c.add_command(register_c)
cloud_c.add_command(curl_c)
cloud_c.add_command(sql_c)
cloud_c.add_command(readme_c)
cloud_c.add_command(description_c)
cloud_c.add_command(metadata_c)
cloud_c.add_command(search_c)
cloud_c.add_command(dump_c)
cloud_c.add_command(load_c)
cloud_c.add_command(token_c)
cloud_c.add_command(add_c)
cloud_c.add_command(status_c)
cloud_c.add_command(logs_c)
cloud_c.add_command(upload_c)
cloud_c.add_command(download_c)
cloud_c.add_command(sync_c)
cloud_c.add_command(plugins_c)
cloud_c.add_command(stub_c)
cloud_c.add_command(validate_c)
cloud_c.add_command(seed_c)



import logging
from abc import ABC, abstractmethod
from hashlib import sha256
from typing import TYPE_CHECKING, Dict, List, Tuple

from splitgraph.core.common import get_temporary_table_id, unmount_schema
from splitgraph.core.engine import lookup_repository
from splitgraph.core.image import Image
from splitgraph.core.repository import Repository, clone
from splitgraph.core.types import ProvenanceLine

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PostgresEngine


def _get_local_image_for_import(hash_or_tag: str, repository: Repository) -> Tuple[Image, bool]:
    """
    Converts a remote repository and tag into an Image object that exists on the engine,
    optionally pulling the repository or cloning it into a temporary location.

    :param hash_or_tag: Hash/tag
    :param repository: Name of the repository (doesn't need to be local)
    :return: Image object and a boolean flag showing whether the repository should be deleted
    when the image is no longer needed.
    """
    tmp_repo = Repository(repository.namespace, repository.repository + "_tmp_clone")
    repo_is_temporary = False

    logging.info("Resolving repository %s", repository)
    source_repo = lookup_repository(repository.to_schema(), include_local=True)
    if source_repo.engine.name != "LOCAL":
        clone(source_repo, local_repository=tmp_repo, download_all=False)
        source_image = tmp_repo.images[hash_or_tag]
        repo_is_temporary = True
    else:
        # For local repositories, first try to pull them to see if they are clones of a remote.
        if source_repo.upstream:
            source_repo.pull(single_image=hash_or_tag)
        source_image = source_repo.images[hash_or_tag]

    return source_image, repo_is_temporary


# TODO: eventually we'll need to reconcile these two classes. They differ in the sequencing and some
#  behaviour:
#  - ImageMapper returns schemas as it gets called; it also clones repos that don't yet exist
#    locally and runs the LQ checkout into a well-known location. It's used by Splitfiles. It can
#    also return provenance for built images.
#  - ImageMounter mounts and returns the schemas in a single bulk call and (the default
#    implementation) puts the schemas in a temporary location, letting checkouts from multiple
#    images co-exist. Used by other transformation plugins, like dbt.


class ImageMapper:
    """
    Image "mapper" used in Splitfile execution that returns the canonical form of an image
    and where it's mounted.
    """

    def __init__(self, object_engine: "PostgresEngine"):
        self.object_engine = object_engine

        self.image_map: Dict[Tuple[Repository, str], Tuple[str, str, Image]] = {}

        self._temporary_repositories: List[Repository] = []

    def _calculate_map(self, repository: Repository, hash_or_tag: str) -> Tuple[str, str, Image]:
        source_image, repo_is_temporary = _get_local_image_for_import(hash_or_tag, repository)
        if repo_is_temporary:
            self._temporary_repositories.append(source_image.repository)

        canonical_form = "%s/%s:%s" % (
            repository.namespace,
            repository.repository,
            source_image.image_hash,
        )

        temporary_schema = sha256(canonical_form.encode("utf-8")).hexdigest()[:63]

        return temporary_schema, canonical_form, source_image

    def __call__(self, repository: Repository, hash_or_tag: str) -> Tuple[str, str]:
        key = (repository, hash_or_tag)
        if key not in self.image_map:
            self.image_map[key] = self._calculate_map(key[0], key[1])

        temporary_schema, canonical_form, _ = self.image_map[key]
        return temporary_schema, canonical_form

    def setup_lq_mounts(self) -> None:
        for temporary_schema, _, source_image in self.image_map.values():
            self.object_engine.delete_schema(temporary_schema)
            self.object_engine.create_schema(temporary_schema)
            source_image.lq_checkout(target_schema=temporary_schema)

    def teardown_lq_mounts(self) -> None:
        for temporary_schema, _, _ in self.image_map.values():
            unmount_schema(self.object_engine, temporary_schema)
        # Delete temporary repositories that were cloned just for the SQL execution.
        for repo in self._temporary_repositories:
            repo.delete()

    def get_provenance_data(self) -> ProvenanceLine:
        return {
            "sources": [
                {
                    "source_namespace": source_repo.namespace,
                    "source": source_repo.repository,
                    "source_hash": source_image.image_hash,
                }
                for (source_repo, _), (_, _, source_image) in self.image_map.items()
            ]
        }


class ImageMounter(ABC):
    """
    Image "mounter" that's used for non-Splitfile transformation jobs (data source plugins that
    implement TransformingDataSource)
    """

    @abstractmethod
    def mount(self, images: List[Tuple[str, str, str]]) -> Dict[Tuple[str, str, str], str]:
        """Mount images into arbitrary schemas and return the image -> schema map"""
        raise NotImplementedError

    @abstractmethod
    def unmount(self) -> None:
        """Unmount the previously mounted images"""
        raise NotImplementedError


class DefaultImageMounter(ImageMounter):
    def __init__(self, engine: "PostgresEngine"):
        self.engine = engine
        self._image_map: Dict[Tuple[str, str, str], str] = {}

    def mount(self, images: List[Tuple[str, str, str]]) -> Dict[Tuple[str, str, str], str]:
        for image in images:
            if image not in self._image_map:
                ns, repo, hash_or_tag = image
                schema = get_temporary_table_id()
                self.engine.delete_schema(schema)
                self.engine.create_schema(schema)
                Repository(ns, repo, engine=self.engine).images[hash_or_tag].lq_checkout(schema)
                self._image_map[image] = schema
        self.engine.commit()
        return self._image_map

    def unmount(self) -> None:
        for tmp_schema in self._image_map.values():
            unmount_schema(self.engine, tmp_schema)
        self._image_map = {}
        self.engine.commit()



import logging
import os
import types
from copy import deepcopy
from importlib import import_module
from typing import Any, Dict, List, Optional, Type, cast

from ...config import CONFIG, get_singleton
from ...config.config import get_all_in_section
from ...config.keys import DEFAULTS
from ...exceptions import DataSourceError
from .base import DataSource
from .fdw import (  # noqa
    ElasticSearchDataSource,
    MongoDataSource,
    MySQLDataSource,
    PostgreSQLDataSource,
)

_DATA_SOURCES: Dict[str, Type[DataSource]] = {}
_data_sources_registered = False


def get_data_source(data_source: str) -> Type[DataSource]:
    """Returns a class for a given data source"""
    global _data_sources_registered
    if not _data_sources_registered:
        _reload_data_sources()
        _data_sources_registered = True

    try:
        return _DATA_SOURCES[data_source]
    except KeyError:
        data_source_class = _load_plugin_from_dir(data_source)
        if data_source_class:
            register_data_source(data_source, data_source_class)
            return data_source_class
        raise DataSourceError("Data source %s not found!" % data_source)


def get_data_sources() -> List[str]:
    """Returns the names of all registered data sources."""
    _reload_data_sources()
    return list(_DATA_SOURCES.keys())


def register_data_source(name: str, data_source_class: Type[DataSource]) -> None:
    """Returns a data source under a given name."""
    global _DATA_SOURCES
    _DATA_SOURCES[name] = data_source_class


def _reload_data_sources() -> None:
    for source_name, source_class_name in get_all_in_section(CONFIG, "data_sources").items():
        assert isinstance(source_class_name, str)

        try:
            data_source = _load_source(source_name, source_class_name)

            assert issubclass(data_source, DataSource)
            register_data_source(source_name.lower(), data_source)
        except (ImportError, AttributeError) as e:
            raise DataSourceError("Error loading custom data source {0}".format(source_name)) from e

    # Load data sources from the additional path
    _register_plugin_dir_data_sources()


def _register_plugin_dir_data_sources():
    plugin_dir = get_singleton(CONFIG, "SG_PLUGIN_DIR")
    if not plugin_dir:
        return
    logging.debug("Looking up plugins in %s", plugin_dir)
    for plugin in os.listdir(plugin_dir):
        data_source = _load_plugin_from_dir(plugin, plugin_dir)
        if data_source:
            register_data_source(plugin, data_source)


def _load_plugin_from_dir(
    plugin_name: str, plugin_dir: Optional[str] = None
) -> Optional[Type[DataSource]]:
    plugin_dir = plugin_dir or get_singleton(CONFIG, "SG_PLUGIN_DIR")

    plugin_file = os.path.join(plugin_dir, plugin_name, "plugin.py")
    if os.path.exists(plugin_file):
        import importlib.util

        # Import the module and get the __plugin__ attr from it
        spec = importlib.util.spec_from_file_location("plugin_dir", plugin_file)
        if not spec:
            logging.warning(
                "Couldn't load the importlib ModuleSpec from %s. Ignoring.", plugin_file
            )
            return None
        module = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(module)  # type:ignore
        if hasattr(module, "__plugin__"):
            logging.debug("Loading %s", plugin_file)
            data_source = module.__plugin__  # type: ignore
            if not issubclass(data_source, DataSource):
                logging.warning(
                    "Data source %s in %s isn't an instance of DataSource. Ignoring.",
                    data_source,
                    plugin_file,
                )
                return None
            return cast(Type[DataSource], data_source)
    return None


def _load_source(source_name, source_class_name):
    # Hack for old-style data sources that have now been moved -- don't crash and instead
    # replace them in the config on the fly
    source_defaults = get_all_in_section(DEFAULTS, "data_sources")

    fallback_used = False

    try:
        ix = source_class_name.rindex(".")
        data_source = getattr(import_module(source_class_name[:ix]), source_class_name[ix + 1 :])
    except (ImportError, AttributeError):
        if source_name not in source_defaults:
            raise
        source_class_name = source_defaults[source_name]
        ix = source_class_name.rindex(".")
        data_source = getattr(import_module(source_class_name[:ix]), source_class_name[ix + 1 :])
        fallback_used = True

    if isinstance(data_source, types.FunctionType):
        if source_name not in source_defaults:
            raise DataSourceError(
                "Handler %s uses the old-style function interface which is not"
                " compatible with this version of Splitgraph. "
                "Delete it from your .sgconfig's [data_sources] section (%s)"
                % (source_name, get_singleton(CONFIG, "SG_CONFIG_FILE"))
            )
        source_class_name = source_defaults[source_name]
        ix = source_class_name.rindex(".")
        data_source = getattr(import_module(source_class_name[:ix]), source_class_name[ix + 1 :])
        fallback_used = True
    if fallback_used:
        logging.warning(
            "Data source %s uses the old-style function interface and was automatically replaced.",
            source_name,
        )
        logging.warning(
            "Replace it with %s=%s in your .sgconfig's [data_sources] section (%s)",
            source_name,
            source_class_name,
            get_singleton(CONFIG, "SG_CONFIG_FILE"),
        )
    return data_source


def merge_jsonschema(left: Dict[str, Any], right: Dict[str, Any]) -> Dict[str, Any]:
    result = deepcopy(left)
    result["properties"] = {**result["properties"], **right.get("properties", {})}
    return result



"""Module imported by Multicorn on the Splitgraph engine server: a foreign data wrapper
that communicates to Socrata datasets using sodapy."""
import json
import logging
from typing import Any, Dict, Optional

import splitgraph.config
from splitgraph.config import get_singleton
from splitgraph.ingestion.socrata.querying import (
    cols_to_socrata,
    estimate_socrata_rows_width,
    quals_to_socrata,
    sortkeys_to_socrata,
)

try:
    from multicorn import ANY, ForeignDataWrapper
except ImportError:
    # Multicorn not installed (OK if we're not on the engine -- tests).
    ForeignDataWrapper = object
    ANY = object()

_PG_LOGLEVEL = logging.INFO


def to_json(row, columns, column_map):
    result = {}
    for col in columns:
        val = row.get(column_map.get(col, col))
        if isinstance(val, (dict, list)):
            val = json.dumps(val)
        result[col] = val
    return result


class SocrataForeignDataWrapper(ForeignDataWrapper):
    def __init__(self, fdw_options, fdw_columns):
        """The foreign data wrapper is initialized on the first query.
        Args:
            fdw_options (dict): The foreign data wrapper options. It is a dictionary
                mapping keys from the sql "CREATE FOREIGN TABLE"
                statement options. It is left to the implementor
                to decide what should be put in those options, and what
                to do with them.

        """
        # Initialize the logger that will log to the engine's stderr: log timestamp and PID.
        from sodapy import Socrata

        logging.basicConfig(
            format="%(asctime)s [%(process)d] %(levelname)s %(message)s",
            level=get_singleton(splitgraph.config.CONFIG, "SG_LOGLEVEL"),
        )

        # Dict of connection parameters as well as the table, repository and image hash to query.
        self.fdw_options = fdw_options

        # The foreign datawrapper columns (name -> ColumnDefinition).
        self.fdw_columns = fdw_columns

        self.table = self.fdw_options["table"]

        # Mappings from SG to Socrata columns (for query building)
        self.column_map = json.loads(self.fdw_options.get("column_map") or "{}")

        self.app_token = self.fdw_options.get("app_token")
        self.domain = self.fdw_options["domain"]
        self.batch_size = int(self.fdw_options.get("batch_size", 1000))
        self.client = Socrata(domain=self.domain, app_token=self.app_token)

        # Cached table metadata
        self._metadata: Optional[Dict[str, Any]] = None

    def can_sort(self, sortkeys):
        """
        :param sortkeys: List of SortKey
        :return: List of SortKey the FDW can sort on
        """

        # Mostly, we can push all sort clauses down to Socrata.
        logging.debug("can_sort %r", sortkeys)

        supported = []
        for key in sortkeys:
            # Socrata sorts nulls first by default (TODO both asc and desc?)
            if key.nulls_first != key.is_reversed:
                continue
            supported.append(key)
        return supported

    def get_rel_size(self, quals, columns):
        """
        Method called from the planner to estimate the resulting relation
        size for a scan.
        It will help the planner in deciding between different types of plans,
        according to their costs.
        Args:
            quals (list): A list of Qual instances describing the filters
                applied to this scan.
            columns (list): The list of columns that must be returned.
        Returns:
            A tuple of the form (expected_number_of_rows, avg_row_width (in bytes))
        """
        try:
            return estimate_socrata_rows_width(columns, self.table_meta, self.column_map)
        except Exception:
            logging.exception("Failed planning Socrata query, returning dummy values")
            return 1000000, len(columns) * 10

    def explain(self, quals, columns, sortkeys=None, verbose=False):
        query = quals_to_socrata(quals, self.column_map)
        select = cols_to_socrata(columns, self.column_map)
        order = sortkeys_to_socrata(sortkeys, self.column_map)

        return [
            "Socrata query to %s" % self.domain,
            "Socrata dataset ID: %s" % self.table,
            "Query: %s" % query,
            "Columns: %s" % select,
            "Order: %s" % order,
        ]

    def execute(self, quals, columns, sortkeys=None):
        """Main Multicorn entry point."""
        query = quals_to_socrata(quals, self.column_map)
        select = cols_to_socrata(columns, self.column_map)
        order = sortkeys_to_socrata(sortkeys, self.column_map)

        logging.debug("Socrata query: %r, select: %r, order: %r", query, select, order)

        # TODO offsets stop working after some point?
        result = self.client.get_all(
            dataset_identifier=self.table,
            where=query,
            select=select,
            limit=self.batch_size,
            order=order,
            exclude_system_fields="false",
        )

        for r in result:
            r = to_json(r, columns, self.column_map)
            yield r

    @property
    def table_meta(self):
        if not self._metadata:
            self._metadata = self.client.get_metadata(dataset_identifier=self.table)
        return self._metadata



"""
Plugin for uploading Splitgraph objects from the cache to an external S3-like object store
"""
import logging
from concurrent.futures import ThreadPoolExecutor
from typing import TYPE_CHECKING, List, Tuple

from tqdm import tqdm

from splitgraph.config import CONFIG, SG_CMD_ASCII, get_singleton
from splitgraph.engine import ResultShape, get_engine
from splitgraph.exceptions import (
    IncompleteObjectDownloadError,
    IncompleteObjectUploadError,
)
from splitgraph.hooks.external_objects import ExternalObjectHandler

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PsycopgEngine

# Downloading/uploading objects to/from S3.
# In the beginning, let's say that we just mount all objects as soon as they are downloaded -- otherwise
# we introduce another distinction between objects that are mounted in splitgraph_meta and objects that
# just exist on a hard drive somewhere.


def get_object_upload_urls(remote_engine, objects):
    remote_engine.run_sql("SET TRANSACTION READ ONLY")
    urls = remote_engine.run_chunked_sql(
        "SELECT splitgraph_api.get_object_upload_urls(%s, %s)",
        ("default", objects),
        return_shape=ResultShape.ONE_ONE,
        chunk_position=1,
    )
    remote_engine.commit()
    remote_engine.close()
    return urls


def get_object_download_urls(remote_engine, remote_object_ids):
    remote_engine.run_sql("SET TRANSACTION READ ONLY")
    urls = remote_engine.run_chunked_sql(
        "SELECT splitgraph_api.get_object_download_urls(%s, %s)",
        ("default", remote_object_ids),
        return_shape=ResultShape.ONE_ONE,
        chunk_position=1,
    )
    remote_engine.commit()
    remote_engine.close()
    return urls


class S3ExternalObjectHandler(ExternalObjectHandler):
    """Uploads/downloads the objects to/from S3/S3-compatible host using the Minio client.

    The handler is "attached" to a given registry which manages issuing pre-signed
    GET/PUT URLs.

    The handler supports a parameter `threads` specifying the number of threads
    used to upload the objects.
    """

    def upload_objects(
        self, objects: List[str], remote_engine: "PsycopgEngine"
    ) -> List[Tuple[str, str]]:
        """
        Upload objects to Minio

        :param remote_engine: Remote Engine class
        :param objects: List of object IDs to upload
        :return: List of tuples with successfully uploaded objects and their URLs.
        """
        worker_threads = self.params.get(
            "threads", int(get_singleton(CONFIG, "SG_ENGINE_POOL")) - 1
        )

        # Determine upload URLs
        logging.info("Getting upload URLs from the registry...")
        urls = get_object_upload_urls(remote_engine, objects)

        local_engine = get_engine()

        def _do_upload(object_url):
            object_id, url = object_url
            # We get 3 URLs here (one for each of object itself, footer and schema -- emit
            # just the first one for logging)
            logging.debug("%s -> %s", object_id, url[0])
            try:
                local_engine.run_api_call("upload_object", object_id, url)
                return object_id
            except Exception:
                logging.exception("Error uploading object %s", object_id)
                return None

        successful: List[str] = []
        try:
            local_engine.autocommit = True
            with ThreadPoolExecutor(max_workers=worker_threads) as tpe:
                pbar = tqdm(
                    tpe.map(_do_upload, zip(objects, urls)),
                    total=len(objects),
                    unit="objs",
                    ascii=SG_CMD_ASCII,
                )
                for object_id in pbar:
                    if object_id:
                        successful.append(object_id)
                        pbar.set_postfix(object=object_id[:10] + "...")
            if len(successful) < len(objects):
                raise IncompleteObjectUploadError(
                    reason=None,
                    successful_objects=successful,
                    successful_object_urls=successful,
                )
            # The "URL" in this case is the same object ID: we ask the registry
            # for the actual URL by giving it the object ID.
            return [(s, s) for s in successful]
        except KeyboardInterrupt as e:
            raise IncompleteObjectUploadError(
                reason=e,
                successful_objects=successful,
                successful_object_urls=successful,
            )
        finally:
            local_engine.autocommit = False
            local_engine.close_others()

    def download_objects(
        self, objects: List[Tuple[str, str]], remote_engine: "PsycopgEngine"
    ) -> List[str]:
        """
        Download objects from Minio.

        :param objects: List of (object ID, object URL (object ID it's stored under))
        """
        # By default, take up the whole connection pool with downloaders
        # (less one connection for the main thread that handles metadata)
        worker_threads = self.params.get(
            "threads", int(get_singleton(CONFIG, "SG_ENGINE_POOL")) - 1
        )

        logging.info("Getting download URLs from registry %s...", remote_engine)
        object_ids = [o[0] for o in objects]
        remote_object_ids = [o[1] for o in objects]
        urls = get_object_download_urls(remote_engine, remote_object_ids)

        local_engine = get_engine()

        def _do_download(object_url):
            object_id, url = object_url
            logging.debug("%s -> %s", url[0], object_id)

            try:
                local_engine.run_api_call("download_object", object_id, url)
                local_engine.mount_object(object_id)
            except Exception as e:
                logging.error("Error downloading object %s: %s", object_id, str(e))

                # Delete the object that we just tried to download to make sure we don't have
                # a situation where the file was downloaded but mounting failed (currently
                # we inspect the filesystem to see the list of downloaded objects).
                # TODO figure out a flow for just remounting objects whose files we already have.
                local_engine.delete_objects([object_id])
                return None

            return object_id

        successful: List[str] = []

        try:
            # Temporarily set the engine into autocommit mode. This is because a transaction
            # commit resets session state and makes the download_object engine API call
            # import all of its Python modules again (which takes about 300ms). It also
            # resets the SD and GD dictionaries so it's not possible to cache those modules
            # there either.
            local_engine.autocommit = True
            with ThreadPoolExecutor(max_workers=worker_threads) as tpe:
                # Evaluate the results so that exceptions thrown by the downloader get raised
                pbar = tqdm(
                    tpe.map(_do_download, zip(object_ids, urls)),
                    total=len(objects),
                    unit="obj",
                    ascii=SG_CMD_ASCII,
                )
                for object_id in pbar:
                    if object_id:
                        successful.append(object_id)
                        pbar.set_postfix(object=object_id[:10] + "...")
            if len(successful) < len(object_ids):
                raise IncompleteObjectDownloadError(reason=None, successful_objects=successful)
            return successful
        except KeyboardInterrupt as e:
            raise IncompleteObjectDownloadError(reason=e, successful_objects=successful)
        finally:
            # Flip the engine back and close all but one pool connection.
            local_engine.autocommit = False
            local_engine.close_others()



"""
Generate a Github Actions workflow file that runs sgr cloud sync on select repositories, respecting
inter-job dependencies.
"""

from typing import Any, Dict, List, Tuple

from splitgraph.cloud.project.utils import get_source_name


def generate_job(
    repository: str, is_live: bool, is_dbt: bool, deploy_url: str = "splitgraph.com"
) -> Tuple[str, Dict[str, Any]]:
    job_id = get_source_name(repository)

    steps = [
        {"uses": "actions/checkout@v2"},
        {
            "name": "Set up Splitgraph",
            "uses": "splitgraph/setup-splitgraph@master",
            "with": {
                "splitgraph_deployment_url": deploy_url,
                "splitgraph_api_key": "${{ secrets.SPLITGRAPH_API_KEY }}",
                "splitgraph_api_secret": "${{ secrets.SPLITGRAPH_API_SECRET }}",
            },
        },
    ]
    if is_dbt:
        steps.append(
            {
                "name": "Set up dbt Git URL",
                "run": 'echo "$CREDENTIALS_YML" > splitgraph.credentials.yml && '
                'sed -i "s|\\$THIS_REPO_URL|https://$GITHUB_ACTOR:$GITHUB_TOKEN@github.com/$GITHUB_REPOSITORY|g" splitgraph.credentials.yml && '
                'sed -i "s|\\$THIS_SHA|$GITHUB_SHA|g" splitgraph.yml',
                "shell": "bash",
                "env": {
                    "CREDENTIALS_YML": "${{secrets.SPLITGRAPH_CREDENTIALS_YML}}",
                    "GITHUB_TOKEN": "${{secrets.GITHUB_TOKEN}}",
                },
            }
        )
    else:
        steps.append(
            {
                "name": "Set up data source credentials",
                "run": 'echo "$CREDENTIALS_YML" > splitgraph.credentials.yml',
                "shell": "bash",
                "env": {"CREDENTIALS_YML": "${{secrets.SPLITGRAPH_CREDENTIALS_YML}}"},
            }
        )

    if is_live:
        # For repositories that support mount, we use `sgr cloud load` to set them up
        # as live data sources and upload metadata.
        steps.append(
            {
                "name": "Run sgr cloud load to set up metadata and data source settings",
                "run": "sgr cloud load --remote splitgraph --initial-private "
                f"-f splitgraph.yml -f splitgraph.credentials.yml {repository}",
                "shell": "bash",
            }
        )
    else:
        # For repositories that don't support mount, first run `sgr cloud sync` to load the data
        # and make the repository private by default. Then, run `sgr cloud load` to set up the
        # repo's README etc (without saving the ingestion settings).
        steps.append(
            {
                "name": "Run sgr cloud sync to perform the data load into a private repo",
                "run": "sgr cloud sync --remote splitgraph --initial-private --use-file "
                f"-f splitgraph.yml -f splitgraph.credentials.yml --wait {repository}",
                "shell": "bash",
            }
        )
        steps.append(
            {
                "name": "Run sgr cloud load to set up metadata",
                "run": "sgr cloud load --remote splitgraph --skip-external "
                f"-f splitgraph.yml -f splitgraph.credentials.yml {repository}",
                "shell": "bash",
            }
        )

    job_doc = {"name": f"Build {repository}", "runs-on": "ubuntu-20.04", "steps": steps}

    return job_id, job_doc


def generate_workflow(
    repositories: List[Tuple[str, bool, bool]], dependencies: Dict[str, List[str]]
) -> Dict[str, Any]:
    job_ids = {}
    job_docs = {}

    for repository, is_live, is_dbt in repositories:
        job_id, job_doc = generate_job(repository, is_live, is_dbt)
        job_ids[repository] = job_id
        job_docs[repository] = job_doc

    for repository, _, _ in repositories:
        repo_deps = dependencies.get(repository)
        if not repo_deps:
            continue

        job_docs[repository]["needs"] = [job_ids[d] for d in repo_deps]

    workflow_doc = {
        "name": "Build datasets on Splitgraph",
        "on": "workflow_dispatch",
        "jobs": {job_ids[r]: job_docs[r] for r, _, _ in repositories},
    }
    return workflow_doc



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['splitgraph',
 'splitgraph.cloud',
 'splitgraph.cloud.project',
 'splitgraph.commandline',
 'splitgraph.config',
 'splitgraph.core',
 'splitgraph.core.indexing',
 'splitgraph.core.sql',
 'splitgraph.engine',
 'splitgraph.engine.postgres',
 'splitgraph.hooks',
 'splitgraph.hooks.data_source',
 'splitgraph.ingestion',
 'splitgraph.ingestion.airbyte',
 'splitgraph.ingestion.csv',
 'splitgraph.ingestion.dbt',
 'splitgraph.ingestion.singer',
 'splitgraph.ingestion.singer.commandline',
 'splitgraph.ingestion.snowflake',
 'splitgraph.ingestion.socrata',
 'splitgraph.resources',
 'splitgraph.resources.icons',
 'splitgraph.resources.splitgraph_meta',
 'splitgraph.resources.static',
 'splitgraph.splitfile',
 'splitgraph.utils']

package_data = \
{'': ['*']}

install_requires = \
['asciitree>=0.3.3',
 'cffi<1.15',
 'chardet>=4.0.0,<5.0.0',
 'click>=7,<8',
 'click_log>=0.3.2',
 'cryptography>=3.4.0',
 'docker>=5.0.2',
 'jsonschema>=3.1.0',
 'minio>=4',
 'packaging>=20.1',
 'parsimonious>=0.8,<0.9',
 'psycopg2-binary>=2,<3',
 'pydantic>=1.8.1',
 'requests>=2.22',
 'ruamel.yaml>=0.17.17,<0.18.0',
 'sodapy>=2.1',
 'splitgraph-pipelinewise-target-postgres>=2.1.0',
 'tabulate>=0.8.7',
 'tqdm>=4.46.0']

extras_require = \
{':sys_platform != "win32"': ['pglast==3.4'],
 'pandas': ['pandas[ingestion]>=0.24', 'sqlalchemy[ingestion]>=1.3,<1.4.23']}

entry_points = \
{'console_scripts': ['sgr = splitgraph.commandline:cli']}

setup_kwargs = {
    'name': 'splitgraph',
    'version': '0.3.6',
    'description': 'Command line library and Python client for Splitgraph, a version control system for data',
    'long_description': '# Splitgraph\n![Build status](https://github.com/splitgraph/splitgraph/workflows/build_all/badge.svg)\n[![Coverage Status](https://coveralls.io/repos/github/splitgraph/splitgraph/badge.svg?branch=master)](https://coveralls.io/github/splitgraph/splitgraph?branch=master)\n[![PyPI version](https://badge.fury.io/py/splitgraph.svg)](https://badge.fury.io/py/splitgraph)\n[![Discord chat room](https://img.shields.io/discord/718534846472912936.svg)](https://discord.gg/4Qe2fYA)\n[![Follow](https://img.shields.io/badge/twitter-@Splitgraph-blue.svg)](https://twitter.com/Splitgraph)\n\n## Overview\n\n**Splitgraph** is a tool for building, versioning and querying reproducible datasets. It\'s inspired\nby Docker and Git, so it feels familiar. And it\'s powered by [PostgreSQL](https://postgresql.org), so it [works seamlessly with existing tools](https://www.splitgraph.com/connect) in the Postgres ecosystem. Use Splitgraph to package your data into self-contained **data images** that you can [share with other Splitgraph instances](https://www.splitgraph.com/docs/getting-started/decentralized-demo).\n\n[**Splitgraph.com**](https://www.splitgraph.com), or **Splitgraph Cloud**, is a public Splitgraph instance where you can share and discover data. It\'s a Splitgraph peer powered by the **Splitgraph Core** code in this repository, adding proprietary features like a data catalog, multitenancy, and a distributed SQL proxy.\n\nYou can explore [40k+ open datasets](https://www.splitgraph.com/explore) in the catalog. You can also connect directly to the [Data Delivery Network](https://www.splitgraph.com/connect) and query any of the datasets, without installing anything.\n\nTo install `sgr` (the command line client) or a local Splitgraph Engine, see the [Installation](#installation) section of this readme.\n\n### Build and Query Versioned, Reproducible Datasets\n\n[**Splitfiles**](https://www.splitgraph.com/docs/concepts/splitfiles) give you a declarative language, inspired by Dockerfiles, for expressing data transformations in ordinary SQL familiar to any researcher or business analyst. You can reference other images, or even other databases, with a simple JOIN.\n\n![](pics/splitfile.png)\n\nWhen you build data with Splitfiles, you get provenance tracking of the resulting data: it\'s possible to find out what sources went into every dataset and know when to rebuild it if the sources ever change. You can easily integrate Splitgraph into your existing CI pipelines, to keep your data up-to-date and stay on top of changes to upstream sources.\n\nSplitgraph images are also version-controlled, and you can manipulate them with Git-like operations through a CLI. You can check out any image into a PostgreSQL schema and interact with it using any PostgreSQL client. Splitgraph will capture your changes to the data, and then you can commit them as delta-compressed changesets that you can package into new images.\n\nSplitgraph supports PostgreSQL [foreign data wrappers](https://wiki.postgresql.org/wiki/Foreign_data_wrappers). We call this feature [mounting](https://www.splitgraph.com/docs/concepts/mounting). With mounting, you can query other databases (like PostgreSQL/MongoDB/MySQL) or open data providers (like [Socrata](https://www.splitgraph.com/docs/ingesting-data/socrata)) from your Splitgraph instance with plain SQL. You can even snapshot the results or use them in Splitfiles.\n\n### Why Splitgraph?\n\nSplitgraph isn\'t opinionated and doesn\'t break existing abstractions. To any existing PostgreSQL application, Splitgraph images are just another database. We have carefully designed Splitgraph to not break the abstraction of a PostgreSQL table and wire protocol, because doing otherwise would mean throwing away a vast existing ecosystem of applications, users, libraries and extensions. This means that a lot of tools that work with PostgreSQL work with Splitgraph out of the box.\n\n![](pics/splitfiles.gif)\n\n## Components\n\nThe code in this repository, known as **Splitgraph Core**, contains:\n\n- **[`sgr` command line client](https://www.splitgraph.com/docs/architecture/sgr-client)**: `sgr` is the main command line tool used to work with Splitgraph "images" (data snapshots). Use it to ingest data, work with splitfiles, and push data to Splitgraph.com.\n- **[Splitgraph Engine](engine/README.md)**: a [Docker image](https://hub.docker.com/r/splitgraph/engine) of the latest Postgres with Splitgraph and other required extensions pre-installed.\n- **[Splitgraph Python library](https://www.splitgraph.com/docs/python-api/splitgraph.core)**: All Splitgraph functionality is available in the Python API, offering first-class support for data science workflows including Jupyter notebooks and Pandas dataframes.\n\n## Docs\n\nDocumentation is available at https://www.splitgraph.com/docs, specifically:\n\n- [Installation](https://www.splitgraph.com/docs/getting-started/installation)\n- [FAQ](https://www.splitgraph.com/docs/getting-started/frequently-asked-questions)\n\nWe also recommend reading our Blog, including some of our favorite posts:\n\n- [Supercharging `dbt` with Splitgraph: versioning, sharing, cross-DB joins](https://www.splitgraph.com/blog/dbt)\n- [Querying 40,000+ datasets with SQL](https://www.splitgraph.com/blog/40k-sql-datasets)\n- [Foreign data wrappers: PostgreSQL\'s secret weapon?](https://www.splitgraph.com/blog/foreign-data-wrappers)\n\n## Installation\n\nPre-requisites:\n\n- Docker is required to run the Splitgraph Engine. `sgr` must have access to Docker. You either need to [install Docker locally](https://docs.docker.com/install/) or have access to a remote Docker socket.\n\nFor Linux and OSX, once Docker is running, install Splitgraph with a single script:\n\n```\n$ bash -c "$(curl -sL https://github.com/splitgraph/splitgraph/releases/latest/download/install.sh)"\n```\n\nThis will download the `sgr` binary and set up the Splitgraph Engine Docker container.\n\nAlternatively, you can get the `sgr` single binary from [the releases page](https://github.com/splitgraph/splitgraph/releases) and run [`sgr engine add`](https://www.splitgraph.com/docs/sgr/engine-management/engine-add) to create an engine.\n\nSee the [installation guide](https://www.splitgraph.com/docs/getting-started/installation) for more installation methods.\n\n## Quick start guide\n\nYou can follow the [quick start guide](https://www.splitgraph.com/docs/getting-started/five-minute-demo) that will guide you through the basics of using Splitgraph with public and private data.\n\nAlternatively, Splitgraph comes with plenty of [examples](examples) to get you started.\n\nIf you\'re stuck or have any questions, check out the [documentation](https://www.splitgraph.com/docs/) or join our [Discord channel](https://discord.gg/4Qe2fYA)!\n\n## Contributing\n\n### Setting up a development environment\n\n  * Splitgraph requires Python 3.6 or later.\n  * Install [Poetry](https://github.com/python-poetry/poetry): `curl -sSL https://raw.githubusercontent.com/python-poetry/poetry/master/get-poetry.py | python` to manage dependencies\n  * Install pre-commit hooks (we use [Black](https://github.com/psf/black) to format code)\n  * `git clone --recurse-submodules https://github.com/splitgraph/splitgraph.git`\n  * `poetry install`\n  * To build the [engine](https://www.splitgraph.com/docs/architecture/splitgraph-engine) Docker image: `cd engine && make`\n\n### Running tests\n\nThe test suite requires [docker-compose](https://github.com/docker/compose). You will also\nneed to add these lines to your `/etc/hosts` or equivalent:\n\n```\n127.0.0.1       local_engine\n127.0.0.1       remote_engine\n127.0.0.1       objectstorage\n```\n\nTo run the core test suite, do\n\n```\ndocker-compose -f test/architecture/docker-compose.core.yml up -d\npoetry run pytest -m "not mounting and not example"\n```\n\nTo run the test suite related to "mounting" and importing data from other databases\n(PostgreSQL, MySQL, Mongo), do\n\n```\ndocker-compose -f test/architecture/docker-compose.core.yml -f test/architecture/docker-compose.mounting.yml up -d\npoetry run pytest -m mounting\n```\n\nFinally, to test the [example projects](https://github.com/splitgraph/splitgraph/tree/master/examples), do\n\n```\n# Example projects spin up their own engines\ndocker-compose -f test/architecture/docker-compose.core.yml -f test/architecture/docker-compose.core.yml down -v\npoetry run pytest -m example\n```\n\nAll of these tests run in [CI](https://github.com/splitgraph/splitgraph/actions).\n',
    'author': 'Splitgraph Limited',
    'author_email': None,
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://www.splitgraph.com',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)



import json
from typing import Callable, Dict, List, Optional, Sequence, Tuple

from splitgraph.core.output import parse_date, parse_dt, parse_time
from splitgraph.core.types import TableColumn, TableSchema
from splitgraph.ingestion.csv.common import pad_csv_row


def parse_boolean(boolean: str):
    if boolean.lower() in ["t", "true"]:
        return True
    elif boolean.lower() in ["f", "false"]:
        return False
    raise ValueError("Invalid boolean!")


def parse_int(integer: str):
    result = int(integer)
    if result > 2147483647 or result < -2147483648:
        raise ValueError("Integer value %s out of range!" % integer)
    return result


def parse_bigint(integer: str):
    result = int(integer)
    if result > 9223372036854775807 or result < -9223372036854775808:
        raise ValueError("Bigint value %s out of range!" % integer)
    return result


def parse_json(json_s: str):
    # Avoid false positives for strings like "  123" that json.loads can
    # parse as JSON but PostgreSQL can't (fall back to strings).
    if "{" not in json_s:
        raise ValueError("Not a JSON object")
    json.loads(json_s)


_CONVERTERS: List[Tuple[str, Callable]] = [
    ("timestamp", parse_dt),
    ("date", parse_date),
    ("time", parse_time),
    ("integer", parse_int),
    ("bigint", parse_bigint),
    ("numeric", float),
    ("boolean", parse_boolean),
    ("json", parse_json),
]


def _infer_column_schema(column_sample: Sequence[str], ignore_empty_strings=True) -> str:
    for candidate, converter in _CONVERTERS:
        try:
            seen_value = False
            for c in column_sample:
                if (c == "" and ignore_empty_strings) or c is None:
                    continue

                seen_value = True
                converter(c)
            # Don't let empty strings or Nones break the parsers but don't accept
            # columns that are just empty strings (they'll be a string).
            if seen_value:
                return candidate
        except (ValueError, TypeError):
            continue

    # No suitable conversion, fall back to varchar
    return "character varying"


def infer_sg_schema(
    sample: Sequence[List[str]],
    override_types: Optional[Dict[str, str]] = None,
    primary_keys: Optional[List[str]] = None,
    ignore_empty_strings: bool = True,
):
    override_types = override_types or {}
    primary_keys = primary_keys or []
    result: TableSchema = []

    header = sample[0]

    sample = [
        pad_csv_row(row, num_cols=len(sample[0]), row_number=i) for i, row in enumerate(sample[1:])
    ]

    columns = list(zip(*sample))
    if len(columns) != len(header):
        raise ValueError(
            "Malformed CSV: header has %d columns, rows have %d columns"
            % (len(header), len(columns))
        )

    for i, (c_name, c_sample) in enumerate(zip(header, columns)):
        pg_type = override_types.get(
            c_name, _infer_column_schema(c_sample, ignore_empty_strings=ignore_empty_strings)
        )

        result.append(
            TableColumn(
                ordinal=i + 1,
                name=c_name,
                pg_type=pg_type,
                is_pk=(c_name in primary_keys),
            )
        )

    return result



import logging
import re
from collections import defaultdict
from datetime import datetime
from typing import (
    TYPE_CHECKING,
    DefaultDict,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    cast,
)

from packaging.version import Version
from psycopg2.sql import SQL, Identifier

from splitgraph.core.sql import insert, select
from splitgraph.engine import ResultShape

if TYPE_CHECKING:
    from splitgraph.engine.postgres.engine import PsycopgEngine

T = TypeVar("T")


def _bfs(adjacency: Dict[T, List[T]], start: T, end: T) -> List[T]:
    queue = [start]
    visited = {start}
    parents: Dict[T, T] = {}

    while queue:
        vertex = queue[0]
        queue = queue[1:]

        if vertex == end:
            path = [end]
            while vertex != start:
                vertex = parents[vertex]
                path.append(vertex)
            path.reverse()
            return path
        for v in adjacency[vertex]:
            if v in visited:
                continue
            visited.add(v)
            parents[v] = vertex
            queue.append(v)

    raise ValueError("No path found between %s and %s" % (start, end))


def make_file_list(schema_name: str, migration_path: List[Optional[str]]):
    """Construct a list of file names from history of versions and schema name"""
    result = []
    for v1, v2 in zip(migration_path, migration_path[1:]):
        if v1 is None:
            name = "%s--%s.sql" % (schema_name, v2)
        else:
            name = "%s--%s--%s.sql" % (schema_name, v1, v2)
        result.append(name)
    return result


def get_installed_version(
    engine: "PsycopgEngine", schema_name: str, version_table: str = "version"
) -> Optional[Tuple[str, datetime]]:
    if not engine.table_exists(schema_name, version_table):
        return None
    return cast(
        Optional[Tuple[str, datetime]],
        engine.run_sql(
            select(version_table, "version,installed", schema=schema_name)
            + SQL("ORDER BY installed DESC LIMIT 1"),
            return_shape=ResultShape.ONE_MANY,
        ),
    )


def _create_version_tracking_schema(
    engine, schema_name: str, version_table: str = "version"
) -> None:
    engine.run_sql(
        SQL(
            """CREATE TABLE {0}.{1} (version VARCHAR NOT NULL,
            installed TIMESTAMP);"""
        ).format(Identifier(schema_name), Identifier(version_table))
    )


def set_installed_version(
    engine: "PsycopgEngine", schema_name: str, version: str, version_table: str = "version"
):
    if not engine.table_exists(schema_name, version_table):
        _create_version_tracking_schema(engine, schema_name, version_table)
    engine.run_sql(
        insert(version_table, ["version", "installed"], schema=schema_name),
        (version, datetime.utcnow()),
    )


def get_version_tuples(filenames: List[str]) -> List[Tuple[Optional[str], str]]:
    version_re = re.compile(r"\w+--([\w.]+)(--[\w.]+)?\.sql")

    result: List[Tuple[Optional[str], str]] = []
    for name in filenames:
        match = version_re.match(name)
        if not match:
            raise ValueError("Invalid file name %s" % name)
        if match.groups()[1] is None:
            result.append((None, match.groups()[0]))
        else:
            result.append((match.groups()[0], match.groups()[1].lstrip("-")))

    return result


def source_files_to_apply(
    engine: "PsycopgEngine",
    schema_name: str,
    schema_files: List[str],
    version_table: str = "version",
    static: bool = False,
    target_version: Optional[str] = None,
) -> Tuple[List[str], str]:
    """Get the ordered list of .sql files to apply to the database"""
    version_tuples = get_version_tuples(schema_files)
    target_version = target_version or max([v[1] for v in version_tuples], key=Version)
    if static:
        # For schemata like splitgraph_api which we want to apply in any
        # case, bypassing the upgrade mechanism, we just run the latest
        # set of sources that this schema needs.
        path = [None, target_version]
    else:
        current = get_installed_version(engine, schema_name, version_table)

        current_version: Optional[str]
        if current:
            current_version, current_install_date = current
            logging.info(
                "Schema %s already exists, version %s, installed on %s",
                schema_name,
                current_version,
                current_install_date,
            )
        else:
            current_version = None

        if current_version == target_version:
            return [], target_version

        # Calculate migration path
        adjacency: DefaultDict[Optional[str], List[Optional[str]]] = defaultdict(list)
        for u, v in version_tuples:
            adjacency[u].append(v)
        path = _bfs(adjacency, start=current_version, end=target_version)

    return make_file_list(schema_name, path), target_version



import logging
from typing import Any, Dict, Generator, Iterable, List, Optional, Tuple

from target_postgres.db_sync import column_type

from splitgraph.config import DEFAULT_CHUNK_SIZE
from splitgraph.core.repository import Repository
from splitgraph.core.types import (
    TableColumn,
    TableInfo,
    TableParams,
    TableSchema,
    get_table_params,
)
from splitgraph.exceptions import TableNotFoundError

from .models import (
    AirbyteCatalog,
    AirbyteMessage,
    AirbyteStream,
    ConfiguredAirbyteCatalog,
    ConfiguredAirbyteStream,
    DestinationSyncMode,
    SyncMode,
)

AirbyteConfig = Dict[str, Any]
AIRBYTE_RAW = "_airbyte_raw"

# By default, the PK on the Airbyte table is _airbyte_ab_id which is a UUID. We also
# cluster by the PK, which means it will be used to detect overlaps and we'll be sorting
# by it, which is bad as currently it forces Splitgraph to materialize a table into PG
# (since it thinks parts of the table overwrite each other). Instead, as a semi-hack,
# we change the table's schema to set the PK as (_airbyte_emitted_at, _airbyte_ab_id)
# which will make it not overlap (as each new chunk will have a different
# _airbyte_emitted_at).
AIRBYTE_RAW_SCHEMA = [
    TableColumn(
        ordinal=1, name="_airbyte_emitted_at", pg_type="timestamp with time zone", is_pk=True
    ),
    TableColumn(ordinal=2, name="_airbyte_ab_id", pg_type="character varying", is_pk=True),
    TableColumn(ordinal=3, name="_airbyte_data", pg_type="jsonb", is_pk=False),
]


def _airbyte_message_reader(
    stream: Iterable[bytes],
) -> Generator[AirbyteMessage, None, None]:
    buffer = b""
    for data in stream:
        # Accumulate data in a buffer until we get a newline, at which point we can
        # decode the message and filter out records/state.
        buffer = buffer + data
        if b"\n" not in data:
            continue

        delimiter = buffer.rindex(b"\n") + 1
        full_message = buffer[:delimiter]
        buffer = buffer[delimiter:]
        lines = full_message.decode().splitlines()

        for line in lines:
            line = line.strip()
            if not line or not line.startswith("{"):
                continue
            message = AirbyteMessage.parse_raw(line)
            yield message


def _store_raw_airbyte_tables(
    repository: Repository,
    image_hash: str,
    staging_schema: str,
    sync_modes: Dict[str, str],
    default_sync_mode: str = "overwrite",
) -> List[str]:
    engine = repository.object_engine
    raw_tables = [t for t in engine.get_all_tables(staging_schema) if t.startswith(AIRBYTE_RAW)]
    current_image = repository.images[image_hash]
    for raw_table in raw_tables:
        sync_mode = sync_modes.get(raw_table)
        if not sync_mode:
            logging.warning(
                "Couldn't detect the sync mode for %s, falling back to %s",
                raw_table,
                default_sync_mode,
            )
            sync_mode = default_sync_mode

        logging.info("Storing %s. Sync mode: %s", raw_table, sync_mode)
        # Make sure the raw table's schema didn't change (very rare, since it's
        # just hash, JSON, timestamp)
        if sync_mode != "overwrite":
            try:
                current_schema = current_image.get_table(raw_table).table_schema
                if current_schema != AIRBYTE_RAW_SCHEMA:
                    raise AssertionError(
                        "Schema for %s changed! Old: %s, new: %s"
                        % (
                            raw_table,
                            current_schema,
                            AIRBYTE_RAW_SCHEMA,
                        )
                    )
            except TableNotFoundError:
                pass

        # If Airbyte meant to overwrite raw tables instead of append to them, we clear out the
        # current raw table so that record_table_as_base doesn't append objects to the existing
        # table.
        if sync_mode == "overwrite":
            repository.objects.overwrite_table(
                repository, image_hash, raw_table, AIRBYTE_RAW_SCHEMA, []
            )

        repository.objects.record_table_as_base(
            repository,
            raw_table,
            image_hash,
            chunk_size=DEFAULT_CHUNK_SIZE,
            source_schema=staging_schema,
            source_table=raw_table,
            table_schema=AIRBYTE_RAW_SCHEMA,
        )

    return raw_tables


def _store_processed_airbyte_tables(
    repository: Repository, image_hash: str, staging_schema: str
) -> None:
    engine = repository.object_engine
    # Save the processed tables in the image
    processed_tables = [
        t for t in engine.get_all_tables(staging_schema) if not t.startswith(AIRBYTE_RAW)
    ]
    for table in processed_tables:
        logging.info("Storing %s", table)
        schema_spec = engine.get_full_table_schema(staging_schema, table)
        repository.objects.overwrite_table(repository, image_hash, table, schema_spec, [])

        repository.objects.record_table_as_base(
            repository,
            table,
            image_hash,
            chunk_size=DEFAULT_CHUNK_SIZE,
            source_schema=staging_schema,
            source_table=table,
        )


def _column_type(schema_property) -> str:
    if "type" not in schema_property:
        # workaround for anyOf
        return "jsonb"
    return str(column_type(schema_property))


def get_sg_schema(stream: AirbyteStream) -> TableSchema:
    # NB Airbyte runs a normalization step after the ingestion that we can't easily predict,
    # since it involves unnesting some fields into separate tables and mapping column names.
    # This is given to the user for informational purposes.
    primary_key = [k for ks in stream.source_defined_primary_key or [] for k in ks]
    return [
        TableColumn(i, name, _column_type(schema_property), name in primary_key, None)
        for i, (name, schema_property) in enumerate(stream.json_schema["properties"].items())
    ]


def get_pk_cursor_fields(
    stream: AirbyteStream,
    table_params: TableParams,
    cursor_overrides: Optional[Dict[str, List[str]]] = None,
    primary_key_overrides: Optional[Dict[str, List[str]]] = None,
) -> Tuple[Optional[List[str]], Optional[List[List[str]]]]:
    cursor_overrides = cursor_overrides or {}
    primary_key_overrides = primary_key_overrides or {}

    # Precedence:
    #  * Override in the table-specific parameters
    #  * Override in the global plugin settings
    #  * Default field in the stream itself (reported by the source)

    cursor_field = stream.default_cursor_field
    custom_cursor_field = table_params.get(
        "airbyte_cursor_field", cursor_overrides.get(stream.name)
    )

    if custom_cursor_field:
        cursor_field = custom_cursor_field

    primary_key = stream.source_defined_primary_key
    custom_primary_key = table_params.get(
        "airbyte_primary_key", primary_key_overrides.get(stream.name)
    )

    if custom_primary_key:
        primary_key = [[k] for k in custom_primary_key]

    return cursor_field, primary_key


def select_streams(
    catalog: AirbyteCatalog,
    tables: Optional[TableInfo],
    sync: bool = False,
    cursor_overrides: Optional[Dict[str, List[str]]] = None,
    primary_key_overrides: Optional[Dict[str, List[str]]] = None,
) -> ConfiguredAirbyteCatalog:
    streams: List[ConfiguredAirbyteStream] = []

    for stream in catalog.streams:
        if tables and stream.name not in tables:
            continue

        cursor_field, primary_key = get_pk_cursor_fields(
            stream,
            get_table_params(tables, stream.name) if tables else TableParams({}),
            cursor_overrides,
            primary_key_overrides,
        )

        # Fix for a weird bug in the file source where the JSONSchema
        # contains "properties" inside of a "properties" object for some reason.
        if "properties" in stream.json_schema["properties"] and isinstance(
            stream.json_schema["properties"]["properties"], dict
        ):
            stream.json_schema["properties"] = stream.json_schema["properties"]["properties"]

        sync_configured = False
        if sync:
            if (
                not stream.supported_sync_modes
                or SyncMode.incremental not in stream.supported_sync_modes
            ):
                logging.warning(
                    "Stream %s doesn't support incremental sync mode and sync=True. "
                    "Disabling append_dedup and falling back to refresh.",
                    stream.name,
                )
            else:
                # Some sources (like google search) issue duplicate fields which breaks mode=append,
                # so we have to use mode=append_dedup. However, that requires an explicit PK which
                # Airbyte currently doesn't extract from Singer-backed sources.
                # PR to fix: https://github.com/airbytehq/airbyte/pull/4789
                # In the meantime, we allow the plugin to override the cursor and the PK field.
                # This is also useful for plugins like Postgres where the user might want to
                # specify their own cursor field.

                if not primary_key or not (cursor_field or stream.source_defined_cursor):
                    logging.warning(
                        "Stream %s doesn't have a primary key or a cursor field/source defined "
                        "cursor (PK: %s, cursor: %s). Disabling append_dedup and falling back "
                        "to refresh.",
                        stream.name,
                        primary_key,
                        cursor_field,
                    )
                else:
                    configured_stream = ConfiguredAirbyteStream(
                        stream=stream,
                        sync_mode=SyncMode.incremental,
                        destination_sync_mode=DestinationSyncMode.append_dedup,
                        cursor_field=cursor_field,
                        primary_key=primary_key,
                    )
                    sync_configured = True

        # Fall back to configuring the stream for full refresh.
        if not sync_configured:
            configured_stream = ConfiguredAirbyteStream(
                stream=stream,
                sync_mode=SyncMode.full_refresh,
                destination_sync_mode=DestinationSyncMode.overwrite,
                cursor_field=cursor_field,
                primary_key=primary_key,
            )

        streams.append(configured_stream)

    return ConfiguredAirbyteCatalog(streams=streams)



"""
Routines related to storing tables as fragments.
"""

import bisect
import itertools
import json
import logging
import operator
import random
import string
import struct
from datetime import datetime
from functools import reduce
from hashlib import sha256
from math import ceil
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Union, cast

from psycopg2._json import Json
from psycopg2.errors import UniqueViolation
from psycopg2.sql import SQL, Composable, Identifier
from tqdm import tqdm

from splitgraph.config import SG_CMD_ASCII, SPLITGRAPH_API_SCHEMA
from splitgraph.core.indexing.bloom import filter_bloom_index, generate_bloom_index
from splitgraph.core.indexing.range import filter_range_index, generate_range_index
from splitgraph.core.metadata_manager import MetadataManager, Object
from splitgraph.core.types import Changeset, TableSchema
from splitgraph.engine import ResultShape, validate_type
from splitgraph.engine.postgres.engine import (
    SG_UD_FLAG,
    add_ud_flag_column,
    chunk,
    get_change_key,
)
from splitgraph.exceptions import SplitGraphError

from .common import SPLITGRAPH_META_SCHEMA, adapt, get_temporary_table_id
from .sql import select

if TYPE_CHECKING:
    from splitgraph.core.repository import Repository
    from splitgraph.core.table import Table
    from splitgraph.engine.postgres.engine import PostgresEngine


def _split_changeset(
    changeset: Changeset, min_max: List[Tuple[Any, Any]], table_pks: List[Tuple[str, str]]
) -> Tuple[List[Changeset], Changeset, Changeset]:
    # maybe order min/max here
    maxs = [m[1] for m in min_max]
    changesets_by_segment: List[Changeset] = [{} for _ in range(len(min_max))]
    before_changesets = {}
    after_changesets = {}
    for pk, data in changeset.items():
        pk = tuple(adapt(v, p[1]) for v, p in zip(pk, table_pks))
        if min_max[0][0] is None or pk < min_max[0][0]:
            before_changesets[pk] = data
            continue
        if pk > min_max[-1][1]:
            after_changesets[pk] = data
            continue
        # This change matches one of the chunks
        changesets_by_segment[bisect.bisect_left(maxs, pk)][pk] = data
    return changesets_by_segment, before_changesets, after_changesets


def _key(c):
    # Hack to work around Nones sometimes being there in comparables. We make a fake
    # key, prepending a boolean (is the item None?) to every item to get a consistent
    # ordering that includes Nones.
    if isinstance(c, tuple):
        return tuple((ci is None, ci) for ci in c)
    return c is None, c


def _pk_overlap(pk_1: Tuple, pk_2: Tuple) -> bool:
    return bool(_key(pk_1[0]) <= _key(pk_2[1]) and _key(pk_1[1]) >= _key(pk_2[0]))


def get_chunk_groups(
    chunks: List[Tuple[str, Any, Any]],
) -> List[List[Tuple[str, Any, Any]]]:
    """
    Takes a list of chunks and their boundaries and combines them
    into independent groups such that chunks from no two groups
    overlap with each other (intervals are assumed to be closed,
    e.g. chunk (1,2) overlaps with chunk (2,3)).

    The original order of chunks is preserved within each group.

    For example, 4 chunks A, B, C, D that don't overlap each other
    will be grouped into 4 groups `[A], [B], [C], [D]`.

    If A overlaps B, the result will be [A, B], [C], [D].

    If in addition B overlaps C (but not A), the result will be `[A, B, C], [D]`.

    If in addition D overlaps any of A, B or C, the result will be `[A, B, C, D]`
    (despite that D is located before A: it will be last since it was last in the
    original list).

    :param chunks: List of (chunk_id, start, end)
    :return: List of lists of (chunk_id, start, end)
    """
    # Note the original order the chunks came in: it should be preserved
    # within overlap groups.

    # no tuple concatenation (typechecker complains)

    chunks: List[Tuple[int, str, Any, Any]] = [
        (i, chunk[0], chunk[1], chunk[2]) for i, chunk in enumerate(chunks)
    ]
    groups: List[List[Tuple[int, str, Any, Any]]] = []
    current_group: List[Tuple[int, str, Any, Any]] = []
    current_group_start = None
    current_group_end = None

    for original_id, chunk_id, start, end in sorted(
        chunks, key=lambda c: _key(c[2])  # type:ignore
    ):
        if not current_group:
            current_group = [(original_id, chunk_id, start, end)]
            current_group_start = start
            current_group_end = end
            continue

        assert current_group_start
        assert current_group_end
        # See if the chunk overlaps with the current chunk group
        if _pk_overlap((start, end), (current_group_start, current_group_end)):
            current_group.append((original_id, chunk_id, start, end))
            current_group_start = min(current_group_start, start, key=_key)
            current_group_end = max(current_group_end, end, key=_key)
            continue

        # If the chunk doesn't overlap, we start a new group
        groups.append(current_group)
        current_group = [(original_id, chunk_id, start, end)]
        current_group_start = start
        current_group_end = end

    groups.append(current_group)

    # Resort the chunks within groups again
    return [[c[1:] for c in sorted(chunks)] for chunks in groups]


class Digest:
    """
    Homomorphic hashing similar to LtHash (but limited to being backed by 256-bit hashes). The main property is that
    for any rows A, B, LtHash(A) + LtHash(B) = LtHash(A+B). This is done by construction: we simply hash individual
    rows and then do bit-wise addition / subtraction of individual hashes to come up with the full table hash.

    Hence, the content hash of any Splitgraph table fragment is the sum of hashes of its added rows minus the sum
    of hashes of its deleted rows (including the old values of the rows that have been updated). This has a very
    useful implication: the hash of a full Splitgraph table is equal to the sum of hashes of its individual fragments.

    This property can be used to simplify deduplication.
    """

    def __init__(self, shorts: Tuple[int, ...]) -> None:
        """
        Create a Digest instance.

        :param shorts: Tuple of 16 2-byte integers.
        """
        # Shorts: tuple of 16 2-byte integers
        assert isinstance(shorts, tuple)
        assert len(shorts) == 16
        self.shorts = shorts

    @classmethod
    def empty(cls) -> "Digest":
        """Return an empty Digest instance such that for any Digest D, D + empty == D - empty == D"""
        return cls((0,) * 16)

    @classmethod
    def from_memoryview(cls, memory: Union[bytes, memoryview]) -> "Digest":
        """Create a Digest from a 256-bit memoryview/bytearray."""
        # Unpack the buffer as 16 signed big-endian shortints.
        return cls(struct.unpack(">16H", memory))

    @classmethod
    def from_hex(cls, hex_string: str) -> "Digest":
        """Create a Digest from a 64-characters (256-bit) hexadecimal string"""
        assert len(hex_string) == 64
        return cls(tuple(int(hex_string[i : i + 4], base=16) for i in range(0, 64, 4)))

    # In these routines, we treat each hash as a vector of 16 2-byte integers and do component-wise addition.
    # To simulate the wraparound behaviour of C shorts, throw away all remaining bits after the action.
    def __add__(self, other: "Digest") -> "Digest":
        return Digest(
            tuple((left + right) & 0xFFFF for left, right in zip(self.shorts, other.shorts))
        )

    def __sub__(self, other: "Digest") -> "Digest":
        return Digest(
            tuple((left - right) & 0xFFFF for left, right in zip(self.shorts, other.shorts))
        )

    def __neg__(self) -> "Digest":
        return Digest(tuple(-v & 0xFFFF for v in self.shorts))

    def hex(self) -> str:
        """Convert the hash into a hexadecimal value."""
        return struct.pack(">16H", *self.shorts).hex()


"""Dictionary of {index_type: [column: index_specific_kwargs or list of columns]}."""
ExtraIndexInfo = Dict[str, Union[List[str], Dict[str, Dict[str, Any]]]]


class FragmentManager(MetadataManager):
    """
    A storage engine for Splitgraph tables. Each table can be stored as one or more immutable fragments that can
    optionally overwrite each other. When a new table is created, it's split up into multiple base fragments. When
    a new version of the table is written, the audit log is inspected and one or more patch fragments are created,
    to be based on the fragments the previous version of the table consisted of. Only the top fragments in this stack
    are stored in the table metadata: to reconstruct the whole table, the links from the top fragments down to the
    base fragments have to be followed.

    In addition, the fragments engine also supports min-max indexing on fragments: this is used to only fetch fragments
    that are required for a given query.
    """

    def __init__(
        self, object_engine: "PostgresEngine", metadata_engine: Optional["PostgresEngine"] = None
    ) -> None:
        metadata_engine = metadata_engine or object_engine
        super().__init__(metadata_engine)
        self.object_engine = object_engine

    def generate_object_index(
        self,
        object_id: str,
        table_schema: TableSchema,
        changeset: Optional[Changeset] = None,
        extra_indexes: Optional[ExtraIndexInfo] = None,
    ) -> Dict[str, Any]:
        """
        Queries the max/min values of a given fragment for each column, used to speed up querying.

        :param object_id: ID of an object
        :param table_schema: Schema of the table the object belongs to.
        :param changeset: Optional, if specified, the old row values are included in the index.
        :param extra_indexes: Dictionary of {index_type: column: index_specific_kwargs}.
        :return: Dict containing the object index.
        """
        extra_indexes: ExtraIndexInfo = extra_indexes or {}

        # Default None, meaning run range index on all columns.
        range_index_columns: Optional[List[str]]
        try:
            range_index_columns = list(extra_indexes["range"])
        except KeyError:
            range_index_columns = None
        range_index: Dict[str, Any] = generate_range_index(
            self.object_engine, object_id, table_schema, changeset, columns=range_index_columns
        )
        indexes = {"range": range_index}

        # Process extra indexes
        for index_name, index_cols in extra_indexes.items():
            if index_name == "range":
                continue
            if index_name != "bloom":
                raise ValueError("Unsupported index type %s!" % index_name)
            if isinstance(index_cols, list):
                raise ValueError(
                    "Unexpected options for index 'bloom': "
                    "got list, expected dictionary {column: {probability/size: ...}}!"
                )

            index_dict = {}
            for index_col, index_kwargs in index_cols.items():
                logging.debug(
                    "Running index %s on column %s with parameters %r",
                    index_name,
                    index_col,
                    index_kwargs,
                )
                index_dict[index_col] = generate_bloom_index(
                    self.object_engine, object_id, changeset, index_col, **index_kwargs
                )
            indexes[index_name] = index_dict

        return indexes

    def _register_object(
        self,
        object_id: str,
        namespace: str,
        insertion_hash: str,
        deletion_hash: str,
        table_schema: TableSchema,
        rows_inserted: int,
        rows_deleted: int,
        changeset: Optional[Changeset] = None,
        extra_indexes: Optional[ExtraIndexInfo] = None,
    ) -> None:
        """
        Registers a Splitgraph object in the object tree and indexes it

        :param object_id: Object ID
        :param namespace: Namespace that owns the object. In registry mode, only namespace owners can alter or delete
            objects.
        :param insertion_hash: Homomorphic hash of all rows inserted by this fragment
        :param deletion_hash: Homomorphic hash of the old values of all rows deleted by this fragment
        :param table_schema: List of (ordinal, name, type, is_pk) with the schema of the table that this object
            belongs to.
        :param changeset: For patches, changeset that produced this object. Must be a dictionary of
            {PK: (True for upserted/False for deleted, old row (if updated or deleted))}. The old values
            are used to generate the min/max index for an object to know if it removes/updates some rows
            that might be pertinent to a query.
        :param extra_indexes: Dictionary of {index_type: column: index_specific_kwargs}.
        """
        object_size = self.object_engine.get_object_size(object_id)
        object_index = self.generate_object_index(object_id, table_schema, changeset, extra_indexes)
        self.register_objects(
            [
                Object(
                    object_id=object_id,
                    format="FRAG",
                    namespace=namespace,
                    size=object_size,
                    created=datetime.utcnow(),
                    insertion_hash=insertion_hash,
                    deletion_hash=deletion_hash,
                    object_index=object_index,
                    rows_inserted=rows_inserted,
                    rows_deleted=rows_deleted,
                )
            ]
        )

    @staticmethod
    def _extract_deleted_rows(changeset: Any, table_schema: TableSchema) -> Any:
        change_key = get_change_key(table_schema)
        pk_cols, _ = zip(*change_key)
        rows = []
        for pk, data in changeset.items():
            if not data[1]:
                # No old row: new value has been inserted.
                continue
            # Turn the changeset into an actual row in the correct order
            pk_index = 0
            row = []
            for col in sorted(table_schema):
                if col.name not in pk_cols:
                    row.append(data[1][col.name])
                else:
                    row.append(pk[pk_index])
                    pk_index += 1
            rows.append(row)
        return rows

    def _hash_old_changeset_values(
        self, changeset: Any, table_schema: TableSchema
    ) -> Tuple[Digest, int]:
        """
        Since we're not storing the values of the deleted rows (and we don't have access to them in the staging table
        because they've been deleted), we have to hash them in Python. This involves mimicking the return value of
        `SELECT t::text FROM table t`.

        :param changeset: Map PK -> (upserted/deleted, Map col -> old val)
        :param table_schema: Table schema
        :return: `Digest` object and the number of deleted rows.
        """
        rows = self._extract_deleted_rows(changeset, table_schema)
        if not rows:
            return Digest.empty(), 0

        # Horror alert: we hash newly created tables by essentially calling digest(row::text) in Postgres and
        # we don't really know how it turns some types to strings. So instead we give Postgres all of its deleted
        # rows back and ask it to hash them for us in the same way.
        inner_tuple = "(" + ",".join("%s::" + validate_type(c.pg_type) for c in table_schema) + ")"
        query = (  # nosec
            "SELECT digest(o::text, 'sha256') FROM (VALUES "
            + ",".join(itertools.repeat(inner_tuple, len(rows)))
            + ") o"
        )

        # By default (e.g. for changesets where nothing was deleted) we use a 0 hash (since adding it to any other
        # hash has no effect).
        digests = self.object_engine.run_sql(
            query,
            [o if not isinstance(o, dict) else Json(o) for row in rows for o in row],
            return_shape=ResultShape.MANY_ONE,
        )
        return (
            reduce(operator.add, map(Digest.from_memoryview, digests), Digest.empty()),
            len(digests),
        )

    def _store_changesets(
        self,
        table: "Table",
        changesets: Any,
        schema: str,
        extra_indexes: Optional[ExtraIndexInfo] = None,
        in_fragment_order: Optional[List[str]] = None,
        overwrite: bool = False,
        table_name: Optional[str] = None,
    ) -> List[str]:
        """
        Store and register multiple changesets as fragments.

        :param table: Table object the changesets belong to
        :param changesets: List of changeset dictionaries. Empty changesets will be ignored.
        :param schema: Schema the table is checked out into.
        :param extra_indexes: Dictionary of {index_type: column: index_specific_kwargs}.
        :param overwrite: Overwrite object if already exists.
        :return: List of created object IDs.
        """
        object_ids = []
        logging.info("Storing and indexing table %s", table.table_name)
        table_name = table_name or table.table_name
        for sub_changeset in tqdm(
            changesets, unit="objs", ascii=SG_CMD_ASCII, disable=len(changesets) < 3
        ):
            if not sub_changeset:
                continue
            # Store the fragment in a temporary location and then find out its hash and rename to the actual target.
            # Optimisation: in the future, we can hash the upserted rows that we need preemptively and possibly
            # avoid storing the object altogether if it's a duplicate.
            tmp_object_id = self._store_changeset(
                sub_changeset, table_name, schema, table.table_schema
            )

            (
                deletion_hash,
                insertion_hash,
                object_id,
                rows_inserted,
                rows_deleted,
            ) = self._get_patch_fragment_hashes_stats(sub_changeset, table, tmp_object_id)

            object_ids.append(object_id)

            # Wrap this rename in a SAVEPOINT so that if the table already exists,
            # the error doesn't roll back the whole transaction (us creating and registering all other objects).
            with self.object_engine.savepoint("object_rename"):
                source_query = SQL("SELECT * FROM {}.{}").format(
                    Identifier("pg_temp"), Identifier(tmp_object_id)
                )

                if in_fragment_order:
                    source_query += SQL(" ") + self._get_order_by_clause(
                        in_fragment_order, table.table_schema
                    )

                try:
                    self.object_engine.store_object(
                        object_id=object_id,
                        source_query=source_query,
                        schema_spec=add_ud_flag_column(table.table_schema),
                        overwrite=overwrite,
                    )
                except UniqueViolation:
                    # Someone registered this object (perhaps a concurrent pull) already.
                    logging.info(
                        "Object %s for table %s/%s already exists, continuing...",
                        object_id,
                        table.repository,
                        table.table_name,
                    )
                self.object_engine.delete_table("pg_temp", tmp_object_id)
                # There are some cases where an object can already exist in the object engine (in the cache)
                # but has been deleted from the metadata engine, so when it's recreated, we'll skip
                # actually registering it. Hence, we still want to proceed trying to register
                # it no matter what.

            # Same here: if we are being called as part of a commit and an object
            # already exists, we'll roll back everything that the caller has done
            # (e.g. registering the new image) if we don't have a savepoint.
            with self.metadata_engine.savepoint("object_register"):
                try:
                    self._register_object(
                        object_id,
                        namespace=table.repository.namespace,
                        insertion_hash=insertion_hash.hex(),
                        deletion_hash=deletion_hash.hex(),
                        table_schema=table.table_schema,
                        changeset=sub_changeset,
                        extra_indexes=extra_indexes,
                        rows_inserted=rows_inserted,
                        rows_deleted=rows_deleted,
                    )
                except UniqueViolation:
                    logging.info(
                        "Object %s for table %s/%s already exists, continuing...",
                        object_id,
                        table.repository,
                        table.table_name,
                    )

        return object_ids

    def _get_patch_fragment_hashes_stats(
        self, sub_changeset: Any, table: "Table", tmp_object_id: str
    ) -> Tuple[Digest, Digest, str, int, int]:
        # Digest the rows.
        deletion_hash, rows_deleted = self._hash_old_changeset_values(
            sub_changeset, table.table_schema
        )
        insertion_hash, rows_inserted = self.calculate_fragment_insertion_hash_stats(
            "pg_temp", tmp_object_id, table.table_schema
        )
        content_hash = (insertion_hash - deletion_hash).hex()
        schema_hash = self._calculate_schema_hash(table.table_schema)
        object_id = "o" + sha256((content_hash + schema_hash).encode("ascii")).hexdigest()[:-2]
        return deletion_hash, insertion_hash, object_id, rows_inserted, rows_deleted

    def _store_changeset(
        self, sub_changeset: Any, table: str, schema: str, table_schema: TableSchema
    ) -> str:
        tmp_object_id = get_temporary_table_id()
        upserted = [pk for pk, data in sub_changeset.items() if data[0]]
        deleted = [pk for pk, data in sub_changeset.items() if not data[0]]
        self.object_engine.store_fragment(
            upserted,
            deleted,
            "pg_temp",
            tmp_object_id,
            schema,
            table,
            table_schema,
        )
        return tmp_object_id

    def calculate_fragment_insertion_hash_stats(
        self, schema: str, table: str, table_schema: Optional[TableSchema] = None
    ) -> Tuple[Digest, int]:
        """
        Calculate the homomorphic hash of just the rows that a given fragment inserts
        :param schema: Schema the fragment is stored in.
        :param table: Name of the table the fragment is stored in.
        :return: A `Digest` object and the number of inserted rows
        """
        table_schema = table_schema or self.object_engine.get_full_table_schema(schema, table)
        columns_sql = SQL(",").join(
            SQL("o.") + Identifier(c.name) for c in table_schema if c.name != SG_UD_FLAG
        )
        digest_query = (
            SQL("SELECT digest((")
            + columns_sql
            + SQL(")::text, 'sha256'::text) FROM {}.{} o").format(
                Identifier(schema), Identifier(table)
            )
            + SQL(" WHERE o.{} = true").format(Identifier(SG_UD_FLAG))
        )
        row_digests = self.object_engine.run_sql(digest_query, return_shape=ResultShape.MANY_ONE)
        insertion_hash = reduce(
            operator.add, (Digest.from_memoryview(r) for r in row_digests), Digest.empty()
        )
        return insertion_hash, len(row_digests)

    def record_table_as_patch(
        self,
        old_table: "Table",
        schema: str,
        image_hash: str,
        new_schema_spec: TableSchema = None,
        split_changeset: bool = False,
        extra_indexes: Optional[ExtraIndexInfo] = None,
        in_fragment_order: Optional[List[str]] = None,
        overwrite: bool = False,
    ) -> None:
        """
        Flushes the pending changes from the audit table for a given table and records them,
        registering the new objects.

        :param old_table: Table object pointing to the current HEAD table
        :param schema: Schema the table is checked out into.
        :param image_hash: Image hash to store the table under
        :param new_schema_spec: New schema of the table (use the old table's schema by default).
        :param split_changeset: See `Repository.commit` for reference
        :param extra_indexes: Dictionary of {index_type: column: index_specific_kwargs}.
        """

        # TODO does the reasoning in the docstring actually make sense? If the point is to, say, for a query
        # (pk=5000) fetch 0 fragments instead of 1 small one, is it worth the cost of having 2 extra fragments?
        # If a changeset is really large (e.g. update the first 1000 rows, update the last 1000 rows) then sure,
        # this will help (for a query pk=5000 we don't need to fetch a 2000-row fragment) but maybe at that point
        # it's time to rewrite the table altogether?

        # Accumulate the diff in-memory. This might become a bottleneck in the future.
        changeset: Changeset = {}
        _conflate_changes(
            changeset,
            cast(
                List[Tuple[Tuple[str, ...], bool, Dict[str, Any], Dict[str, Any]]],
                self.object_engine.get_pending_changes(schema, old_table.table_name),
            ),
        )
        self.object_engine.discard_pending_changes(schema, old_table.table_name)
        current_objects = old_table.objects

        new_schema_spec = new_schema_spec or old_table.table_schema
        if changeset:
            if split_changeset:
                logging.debug("Splitting changesets")
                # Reorganize the current table's fragments into non-overlapping groups
                # and split the changeset to make sure it doesn't span (and hence merge) them.
                table_pks = self.object_engine.get_change_key(schema, old_table.table_name)
                changesets = self.split_changeset_boundaries(changeset, table_pks, current_objects)
            else:
                changesets = [changeset]

            # Store the changesets and find out their object IDs.
            object_ids = self._store_changesets(
                old_table,
                changesets,
                schema,
                extra_indexes,
                in_fragment_order=in_fragment_order,
                overwrite=overwrite,
            )
            # Finally, link the table to the new set of objects.
            self.register_tables(
                old_table.repository,
                [
                    (
                        image_hash,
                        old_table.table_name,
                        new_schema_spec,
                        old_table.objects + object_ids,
                    )
                ],
            )
        else:
            # Changes in the audit log cancelled each other out. Point the image to the same old objects.
            self.register_tables(
                old_table.repository,
                [(image_hash, old_table.table_name, new_schema_spec, old_table.objects)],
            )

    def split_changeset_boundaries(
        self, changeset: Changeset, change_key: List[Tuple[str, str]], objects: List[str]
    ) -> List[Changeset]:
        if not objects:
            return [changeset]

        min_max = self.get_min_max_pks(objects, change_key)
        groups = get_chunk_groups([(o, mm[0], mm[1]) for o, mm in zip(objects, min_max)])
        group_boundaries = [
            (min(min_pk for _, min_pk, _ in group), max(max_pk for _, _, max_pk in group))
            for group in groups
        ]
        matched, before, after = _split_changeset(changeset, group_boundaries, change_key)
        changesets = [before] + matched + [after]
        return changesets

    def get_min_max_pks(
        self, fragments: List[str], table_pks: List[Tuple[str, str]]
    ) -> List[Tuple[Tuple, Tuple]]:
        """Get PK ranges for given fragments using the index (without reading the fragments).

        :param fragments: List of object IDs (must be registered and with the same schema)
        :param table_pks: List of tuples (column, type) that form the object PK.

        :return: List of (min, max) PK for every fragment where PK is a tuple.
            If a fragment doesn't exist or doesn't have a corresponding index entry,
            a SplitGraphError is raised.
        """
        # If the PK isn't composite, we can read the range for the corresponding column
        # from the index, otherwise, the indexer stored the min/max tuple under $pk.
        pk = table_pks[0][0] if len(table_pks) == 1 else "$pk"
        fields = SQL("object_id, index #>> '{{range,{0},0}}', index #>> '{{range,{0},1}}'").format(
            Identifier(pk)
        )

        result = {
            r[0]: (r[1], r[2])
            for r in self.metadata_engine.run_chunked_sql(
                select(
                    "get_object_meta",
                    fields.as_string(self.metadata_engine.connection),
                    table_args="(%s)",
                    schema=SPLITGRAPH_API_SCHEMA,
                ),
                (fragments,),
                chunk_position=0,
            )
        }

        # Since the PK can't contain a NULL, if we do get one here, it's from the JSON query
        # (column doesn't exist in the index).

        min_max = []
        for fragment in fragments:
            if fragment not in result:
                raise SplitGraphError("No metadata found for object %s!" % fragment)
            min_pk, max_pk = result[fragment]
            if min_pk is None or max_pk is None:
                raise SplitGraphError("No index found for object %s!" % fragment)
            if pk == "$pk":
                # For composite PKs, we're given back a JSON array and need to load it.
                min_pk = tuple(json.loads(min_pk))
                max_pk = tuple(json.loads(max_pk))
            else:
                # Single-column PKs still need to be returned as tuples.
                min_pk = (min_pk,)
                max_pk = (max_pk,)

            # Coerce the PKs to the actual Python types
            min_pk = tuple(adapt(v, c[1]) for v, c in zip(min_pk, table_pks))
            max_pk = tuple(adapt(v, c[1]) for v, c in zip(max_pk, table_pks))
            min_max.append((min_pk, max_pk))

        return min_max

    def calculate_content_hash(
        self,
        schema: str,
        table: str,
        table_schema: Optional[TableSchema] = None,
        chunk_condition_sql: Optional[Composable] = None,
        chunk_condition_args: Optional[List[Any]] = None,
    ) -> Tuple[str, int]:
        """
        Calculates the homomorphic hash of table contents.

        :param schema: Schema the table belongs to
        :param table: Name of the table
        :param table_schema: Schema of the table
        :param chunk_condition_sql: Column the table is partitioned on
        :param chunk_condition_args: Column value to get rows from
        :return: A 64-character (256-bit) hexadecimal string with the content hash of the table
            and the number of rows in the hash.
        """
        table_schema = table_schema or self.object_engine.get_full_table_schema(schema, table)
        digest_query = (
            SQL("SELECT digest((")
            + SQL(",").join(Identifier(c.name) for c in table_schema)
            + SQL(")::text, 'sha256'::text) FROM {}.{} o").format(
                Identifier(schema), Identifier(table)
            )
        )
        args = None
        if chunk_condition_sql:
            digest_query += SQL(" ") + chunk_condition_sql
            args = chunk_condition_args

        row_digests = self.object_engine.run_sql(
            digest_query, args, return_shape=ResultShape.MANY_ONE
        )

        return (
            reduce(operator.add, (Digest.from_memoryview(r) for r in row_digests)).hex(),
            len(row_digests),
        )

    def create_base_fragment(
        self,
        source_schema: Optional[str],
        source_table: Optional[str],
        namespace: str,
        table_schema: TableSchema,
        source_function: Optional[Composable] = None,
        extra_indexes: Optional[ExtraIndexInfo] = None,
        in_fragment_order: Optional[List[str]] = None,
        overwrite: bool = False,
    ) -> str:
        if not source_schema and not source_table and not source_function:
            raise ValueError("Pass either source_schema/table or source_function!")
        schema_hash = self._calculate_schema_hash(table_schema)

        # Object IDs are also used to key tables in Postgres so they can't be more than 63 characters.
        # In addition, table names can't start with a number (they can but every invocation has to
        # be quoted) so we have to drop 2 characters from the 64-character hash and append an "o".

        # Store the object adding the extra update/delete column (always True in this case
        # since we don't overwrite any rows) and filtering on the chunk ID.

        source_query = (
            SQL("SELECT ")
            + SQL(",").join(Identifier(c.name) for c in table_schema)
            + SQL(",TRUE AS ")
            + Identifier(SG_UD_FLAG)
        )
        if source_function:
            source_query += SQL(" FROM ") + source_function
        else:
            source_query += SQL("FROM {}.{}").format(
                Identifier(source_schema), Identifier(source_table)
            )
        source_query_args = None

        if in_fragment_order:
            source_query += SQL(" ") + self._get_order_by_clause(in_fragment_order, table_schema)

        tmp_object_id = get_temporary_table_id()
        self.object_engine.store_object(
            object_id=tmp_object_id,
            source_query=source_query,
            schema_spec=add_ud_flag_column(table_schema),
            source_query_args=source_query_args,
            overwrite=overwrite,
        )

        # Get content hash for this chunk.
        content_hash, rows_inserted = self.calculate_content_hash(
            SPLITGRAPH_META_SCHEMA,
            tmp_object_id,
            table_schema,
        )
        object_id = "o" + sha256((content_hash + schema_hash).encode("ascii")).hexdigest()[:-2]
        try:
            # Check if the object metadata and the object files exist. Sometimes we might have
            # one but not the other, in which case we want to get out of this invalid state.
            if (
                self.get_object_meta([object_id])
                and self.object_engine.run_api_call("object_exists", object_id)
                and not overwrite
            ):
                logging.info(
                    "Object %s already exists, continuing...",
                    object_id,
                )
                return object_id

            # Rename the object to its actual ID, index and register it.

            # What if the object doesn't exist before the previous check but gets created on
            # this line by a concurrent pull? The next two routines have upsert semantics. In the
            # rare case of a race condition, we'll rename the existing object and overwrite it
            # with the same file (since the hash is the same), then register it, overwriting object
            # metadata with the same metadata.

            self.object_engine.rename_object(
                old_object_id=tmp_object_id,
                new_object_id=object_id,
            )
            self._register_object(
                object_id,
                namespace=namespace,
                insertion_hash=content_hash,
                deletion_hash="0" * 64,
                table_schema=table_schema,
                extra_indexes=extra_indexes,
                rows_inserted=rows_inserted,
                rows_deleted=0,
            )
        finally:
            self.object_engine.delete_objects([tmp_object_id])

        return object_id

    @staticmethod
    def _get_order_by_clause(in_fragment_order, table_schema):
        column_names = [s.name for s in table_schema]
        if in_fragment_order:
            for c in in_fragment_order:
                if c not in column_names:
                    raise ValueError("Unknown column name %s, can't sort by it!" % c)
        return SQL("ORDER BY ") + SQL(",").join(Identifier(c) for c in in_fragment_order)

    @staticmethod
    def _calculate_schema_hash(table_schema):
        # Don't include column comments in the schema hash.
        return sha256(
            str([(c.ordinal, c.name, c.pg_type, c.is_pk) for c in table_schema]).encode("utf-8")
        ).hexdigest()

    def record_table_as_base(
        self,
        repository: "Repository",
        table_name: str,
        image_hash: str,
        chunk_size: Optional[int] = 10000,
        source_schema: Optional[str] = None,
        source_table: Optional[str] = None,
        extra_indexes: Optional[ExtraIndexInfo] = None,
        in_fragment_order: Optional[List[str]] = None,
        overwrite: bool = False,
        table_schema: Optional[TableSchema] = None,
    ) -> List[str]:
        """
        Copies the full table verbatim into one or more new base fragments and registers them.

        :param repository: Repository
        :param table_name: Table name
        :param image_hash: Hash of the new image
        :param chunk_size: If specified, splits the table into multiple objects with a given number of rows
        :param source_schema: Override the schema the source table is stored in
        :param source_table: Override the name of the table the source is stored in
        :param extra_indexes: Dictionary of {index_type: column: index_specific_kwargs}.
        :param in_fragment_order: Key to sort data inside each chunk by.
        :param overwrite: Overwrite physical objects that already exist.
        :param table_schema: Override the columns that will be picked from the original table
            (e.g. to change their order or primary keys). Note that the schema must be a subset
            of the original schema and this method doesn't verify PK constraints.
        """
        source_schema = source_schema or repository.to_schema()
        source_table = source_table or table_name

        table_not_empty = self.object_engine.run_sql(
            SQL("SELECT EXISTS(SELECT 1 FROM {}.{})").format(
                Identifier(source_schema), Identifier(source_table)
            ),
            return_shape=ResultShape.ONE_ONE,
        )

        table_schema = table_schema or self.object_engine.get_full_table_schema(
            source_schema, source_table
        )
        if chunk_size and table_not_empty:
            object_ids = self._chunk_table(
                repository,
                source_schema,
                source_table,
                chunk_size,
                extra_indexes,
                in_fragment_order=in_fragment_order,
                overwrite=overwrite,
                table_schema=table_schema,
            )

        elif table_not_empty:
            object_ids = [
                self.create_base_fragment(
                    source_schema,
                    source_table,
                    repository.namespace,
                    extra_indexes=extra_indexes,
                    in_fragment_order=in_fragment_order,
                    overwrite=overwrite,
                    table_schema=table_schema,
                )
            ]
        else:
            # If the table is empty, then we don't link it to any objects and simply store its schema
            object_ids = []
        self.register_tables(repository, [(image_hash, table_name, table_schema, object_ids)])
        return object_ids

    def _chunk_table(
        self,
        repository: "Repository",
        source_schema: str,
        source_table: str,
        chunk_size: int,
        extra_indexes: Optional[ExtraIndexInfo] = None,
        table_schema: Optional[TableSchema] = None,
        in_fragment_order: Optional[List[str]] = None,
        overwrite: bool = False,
    ) -> List[str]:
        table_schema = table_schema or self.object_engine.get_full_table_schema(
            source_schema, source_table
        )
        change_key = get_change_key(table_schema)
        table_pk = [p[0] for p in change_key]
        surrogate_pk = not any(t.is_pk for t in table_schema)

        # We need to do multiple things here in a specific way to not tank the performance:
        #  * Chunk the table up ordering by PK (or potentially another chunk key in the future)
        #  * Run LTHash on added rows in every chunk
        #  * Copy each chunk into a CStore table (adding ranges/bloom filter intex to it
        #    in the metadata).
        #
        # Chunking is very slow to do with
        #   SELECT * FROM source LIMIT chunk_size OFFSET offset ORDER BY pk
        # as Postgres needs (even if there's an index on the PK) to go through first `offset` tuples
        # before finding out what it is it's supposed to copy. So for the full table, PG will do
        # 0 + chunk_size + 2 * chunk_size + ... + (no_chunks - 1) * chunk_size fetches
        # which is O(n^2).
        #
        # The second strategy here was recording the last PK we saw and then copying the next
        # fragment out of the table with SELECT ... WHERE pk > last_pk ORDER BY pk LIMIT chunk_size
        # but that still led to poor performance on large tables (8M rows, chunks of ~200k rows
        # would take 1 minute each to create).
        #
        # Third attempt was adding a temporary column to the source table using RANK () OVER
        #   (ORDER BY pk) but that required a join with a CTE and a couple of sequential
        # scans which also took more than 15 minutes on a 8M row table, no matter whether the
        # table had indexes on the join key.
        #
        # Fourth attempt was: compute the partition key and extract the table contents
        # into a TEMPORARY table, then create an index on that partition key, then copy data
        # out of it into CStore. This still meant having to copy the table over twice.
        #
        # Fifth attempt: create a temporary table with partition boundaries and then
        # use that as the chunking condition (lower <= pk < upper). This led to poor performance
        # for tables without a primary key, as well as foreign tables (lots of round trips).
        #
        # Current attempt: using a named cursor and FETCHing a certain number of rows into separate
        # cstore_fdw tables to get this done in a single table scan.

        logging.info("Processing table %s", source_table)
        pk_sql = SQL(",").join(Identifier(p) for p in table_pk)

        # If we don't have a PK, we cast the whole row to text when partitioning
        # the table as well as selecting from it
        cast_if_surrogate = SQL("::text" if surrogate_pk else "")
        object_ids = []

        table_length = self.object_engine.run_sql(
            SQL("SELECT COUNT(1) FROM {}.{}").format(
                Identifier(source_schema), Identifier(source_table)
            ),
            return_shape=ResultShape.ONE_ONE,
        )
        total_chunks = int(ceil(table_length / chunk_size))
        log_progress = total_chunks > 10
        log_func = logging.info if log_progress else logging.debug

        log_func("Storing and indexing the table")

        # Declare a cursor that will read from the source table (with an order)
        cursor_name = "_sg_commit_" + "".join(
            random.choice(string.ascii_lowercase + string.ascii_uppercase + string.digits)
            for _ in range(10)
        )
        cursor_query = (
            SQL("DECLARE {} CURSOR FOR (SELECT").format(Identifier(cursor_name))
            + SQL(",").join(Identifier(c.name) for c in table_schema)
            + SQL("FROM {}.{}").format(Identifier(source_schema), Identifier(source_table))
            + SQL(" ORDER BY (")
            + pk_sql
            + SQL(")")
            + cast_if_surrogate
            + SQL(");")
        )

        # Giant hack to work around the fact that we can't FETCH directly into a table.
        # We create a function that returns a table with the same shape as ours, then get
        # it to run the FETCH for us instead and call that.
        # Thanks to https://stackoverflow.com/a/50855367
        cursor_query += (
            SQL("CREATE OR REPLACE FUNCTION {}._sg_tmp_read_cursor() RETURNS TABLE(").format(
                Identifier(source_schema)
            )
            + SQL(",".join("{} %s " % validate_type(col.pg_type) for col in table_schema)).format(
                *(Identifier(col.name) for col in table_schema)
            )
            + SQL(
                ") AS 'BEGIN RETURN QUERY FETCH "
                + str(int(chunk_size))
                + " FROM {}; END' LANGUAGE plpgsql"
            ).format(Identifier(cursor_name))
        )
        try:
            self.object_engine.run_sql(cursor_query)

            pbar = tqdm(
                range(total_chunks), unit="objs", ascii=SG_CMD_ASCII, disable=not log_progress
            )
            for _ in pbar:
                object_id = self.create_base_fragment(
                    None,
                    None,
                    repository.namespace,
                    source_function=SQL("{}._sg_tmp_read_cursor()").format(
                        Identifier(source_schema)
                    ),
                    extra_indexes=extra_indexes,
                    in_fragment_order=in_fragment_order,
                    overwrite=overwrite,
                    table_schema=table_schema,
                )

                object_ids.append(object_id)
                pbar.set_postfix(object=object_id[:10] + "...")
        finally:
            self.object_engine.run_sql(
                SQL("DROP FUNCTION IF EXISTS {}._sg_tmp_read_cursor()").format(
                    Identifier(source_schema)
                )
            )
            # Drop the cursor if it exists
            if self.object_engine.run_sql(
                "SELECT 1 FROM pg_cursors WHERE name = %s", (cursor_name,)
            ):
                self.object_engine.run_sql(SQL("CLOSE ") + Identifier(cursor_name))
        return object_ids

    def filter_fragments(self, object_ids: List[str], table: "Table", quals: Any) -> List[str]:
        """
        Performs fuzzy filtering on the given object IDs using the index and a set of qualifiers, discarding
        objects that definitely do not match the qualifiers.

        :param object_ids: List of object IDs to filter.
        :param table: A Table object the objects belong to.
        :param quals: List of qualifiers in conjunctive normal form that will be matched against the index.
            Objects that definitely don't match these qualifiers will be discarded.

            A list containing `[[qual_1, qual_2], [qual_3, qual_4]]` will be interpreted as
            (qual_1 OR qual_2) AND (qual_3 OR qual_4).

            Each qual is a tuple of `(column_name, operator, value)` where
            `operator` can be one of `>`, `>=`, `<`, `<=`, `=`.

            For unknown operators, it will be assumed that all fragments might match that clause.
        :return: List of objects that might match the given qualifiers.
        """
        if not quals:
            return object_ids

        column_types = {c[1]: c[2] for c in table.table_schema}

        # Run the range filter
        range_filter_result = filter_range_index(
            self.metadata_engine, object_ids, quals, column_types
        )
        if len(range_filter_result) < len(object_ids):
            logging.info(
                "Range filter discarded %d/%d fragment(s)",
                len(object_ids) - len(range_filter_result),
                len(object_ids),
            )

        # Run other filters: currently we can attempt to run the bloom filter
        # if the fragment metadata has bloom fingerprints.
        bloom_filter_result = filter_bloom_index(self.metadata_engine, range_filter_result, quals)
        if len(bloom_filter_result) < len(range_filter_result):
            logging.info(
                "Bloom filter discarded %d/%d fragment(s)",
                len(range_filter_result) - len(bloom_filter_result),
                len(range_filter_result),
            )

        objects_to_scan = self._add_overlapping_objects(table, object_ids, bloom_filter_result)

        if len(objects_to_scan) > len(bloom_filter_result):
            logging.info(
                "Will need to scan through extra %d overlapping fragment(s)",
                len(objects_to_scan) - len(bloom_filter_result),
            )

        # Preserve original object order.
        return [r for r in object_ids if r in objects_to_scan]

    def generate_surrogate_pk(
        self, table: "Table", object_pks: List[Tuple[Any, Any]]
    ) -> List[Tuple[Any, Any]]:
        """
        When partitioning data, if the table doesn't have a primary key, we use a "surrogate"
        primary key by concatenating the whole row as a string on the PG side (this is because
        the whole row can sometimes contain NULLs which we can't compare in PG).

        We need to mimic this when calculating if the objects we're about to scan through
        overlap: e.g. using string comparison, "(some_country, 100)" < "(some_country, 20)",
        whereas using typed comparison, (some_country, 100) > (some_country, 20).

        To do this, we use a similar hack from when calculating changeset hashes: to avoid having
        to reproduce how PG's ::text works, we give it back the rows and get it to cast them
        to text for us.
        """
        inner_tuple = (
            "(" + ",".join("%s::" + ct for _, ct in get_change_key(table.table_schema)) + ")"
        )
        rows = [r for o in object_pks for r in o]

        result = []
        for batch in chunk(rows, 1000):
            query = (  # nosec
                "SELECT o::text FROM (VALUES "
                + ",".join(itertools.repeat(inner_tuple, len(batch)))
                + ") o"
            )
            result.extend(
                self.object_engine.run_sql(
                    query,
                    [o if not isinstance(o, dict) else Json(o) for row in batch for o in row],
                    return_shape=ResultShape.MANY_ONE,
                )
            )
        object_pks = list(zip(result[::2], result[1::2]))
        return object_pks

    def _add_overlapping_objects(
        self, table: "Table", all_objects: List[str], filtered_objects: List[str]
    ) -> Set[str]:
        # Expand the list of objects by also adding the objects that might overwrite these. In some
        # cases, we don't keep track of the rows that an object deletes in the index, since that
        # adds an implicit dependency on those previous objects.

        table_pk = get_change_key(table.table_schema)
        object_pks = self.get_min_max_pks(all_objects, table_pk)

        surrogate_pk = not any(t.is_pk for t in table.table_schema)
        if surrogate_pk:
            object_pks = self.generate_surrogate_pk(table, object_pks)

        # Go through all objects and see if they 1) come after any of our chosen objects and 2)
        # overlap those objects' PKs (if they come after them)
        original_order = {object_id: i for i, object_id in enumerate(all_objects)}
        object_pk_dict = {
            object_id: object_pk for object_id, object_pk in zip(all_objects, object_pks)
        }
        objects_to_scan = set(filtered_objects)
        for overlap_candidate in all_objects:
            if overlap_candidate in objects_to_scan:
                continue
            for our_object in filtered_objects:
                if original_order[our_object] < original_order[overlap_candidate] and _pk_overlap(
                    object_pk_dict[our_object], object_pk_dict[overlap_candidate]
                ):
                    objects_to_scan.add(overlap_candidate)
                    break
        return objects_to_scan

    def delete_objects(self, objects: Union[Set[str], List[str]]) -> None:
        """
        Deletes objects from the Splitgraph cache

        :param objects: A sequence of objects to be deleted
        """
        objects = list(objects)
        for i in range(0, len(objects), 100):
            to_delete = objects[i : i + 100]
            table_types = self.object_engine.run_sql(
                SQL(  # nosec
                    "SELECT table_name, table_type FROM information_schema.tables "
                    "WHERE table_schema = %s AND table_name IN ("
                    + ",".join(itertools.repeat("%s", len(to_delete)))
                    + ")"
                ),
                [SPLITGRAPH_META_SCHEMA] + to_delete,
            )

            base_tables = [tn for tn, tt in table_types if tt == "BASE TABLE"]
            # Try deleting CStore-mounted objects regardless of whether they're
            # in splitgraph_meta as foreign tables (there might be cases
            # where they are in /var/lib/splitgraph/objects but not mounted)
            foreign_tables = [tn for tn in to_delete if tn not in base_tables]

            if base_tables:
                self.object_engine.run_sql(
                    SQL(";").join(
                        SQL("DROP TABLE IF EXISTS {}.{}").format(
                            Identifier(SPLITGRAPH_META_SCHEMA), Identifier(t)
                        )
                        for t in base_tables
                    )
                )
            if foreign_tables:
                self.object_engine.delete_objects(foreign_tables)

            self.object_engine.commit()


def _conflate_changes(
    changeset: Changeset, new_changes: List[Tuple[Tuple, bool, Dict[str, Any], Dict[str, Any]]]
) -> Changeset:
    """
    Updates a changeset to incorporate the new changes. Assumes that the new changes are non-pk changing
    (i.e. PK-changing updates have been converted into a del + ins).
    """
    for change_pk, upserted, old_row, new_row in new_changes:
        old_change = changeset.get(change_pk)
        if not old_change:
            changeset[change_pk] = (upserted, old_row, new_row)
        else:
            # If we reinserted the same row that we deleted or deleted a row that
            # was inserted, drop the change completely.
            if old_change[1] == new_row:
                del changeset[change_pk]
            else:
                changeset[change_pk] = (upserted, old_change[1], new_row)

    return changeset
